# pyflyby/_parse.py.
# Copyright (C) 2011, 2012, 2013, 2014, 2015, 2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT
from __future__ import annotations, print_function

import ast
from   ast                      import AsyncFunctionDef, TypeIgnore

from   collections              import namedtuple
from   doctest                  import DocTestParser
from   functools                import cached_property, total_ordering
from   itertools                import groupby

from   pyflyby._file            import FilePos, FileText, Filename
from   pyflyby._flags           import CompilerFlags
from   pyflyby._log             import logger
from   pyflyby._util            import cmp

import re
import sys
from   textwrap                 import dedent
import types
from   typing                   import Any, List, Optional, Tuple, Union, cast, Literal
import warnings


_sentinel = object()

if sys.version_info < (3, 10):

    NoneType = type(None)

    class MatchAs:
        name: str
        pattern: ast.AST

    class MatchMapping:
        keys: List[ast.AST]
        patterns: List[MatchAs]

else:
    from types import NoneType
    from ast import MatchAs, MatchMapping


def _is_comment_or_blank(line, /):
    """
    Returns whether a line of python code contains only a comment is blank.

      >>> _is_comment_or_blank("foo\\n")
      False

      >>> _is_comment_or_blank("  # blah\\n")
      True
    """
    return re.sub("#.*", "", line).rstrip() == ""


def _is_ast_str_or_byte(node) -> bool:
    """
    utility function that test if node is an ast.Str|ast.Bytes in Python < 3.12,
    and if it is a ast.Constant, with node.value being a str in newer version.
    """
    return _is_ast_str(node) or _is_ast_bytes(node)

def _is_ast_bytes(node) -> bool:
    """
    utility function that test if node is an ast.Str in Python < 3.12,
    and if it is a ast.Constant, with node.value being a str in newer version.
    """
    if sys.version_info < (3,12):
        return isinstance(node, ast.Bytes)
    else:
        return (isinstance(node, ast.Constant) and isinstance(node.value , bytes))


def _is_ast_str(node) -> bool:
    """
    utility function that test if node is an ast.Str in Python < 3.12,
    and if it is a ast.Constant, with node.value being a str in newer version.
    """
    if sys.version_info < (3,12):
        return isinstance(node, ast.Str)
    else:
        return (isinstance(node, ast.Constant) and isinstance(node.value , str))

def _ast_str_literal_value(node):
    if _is_ast_str_or_byte(node):
        return node.s
    if isinstance(node, ast.Expr) and _is_ast_str_or_byte(node.value):
        if sys.version_info > (3,10):
            return node.value.value
        else:
            return node.value.s
    else:
        return None


def _flatten_ast_nodes(arg):
    if arg is None:
        pass
    elif isinstance(arg, ast.AST):
        yield arg
    elif isinstance(arg, str):
        #FunctionDef type_comments
        yield arg
    elif isinstance(arg, (tuple, list, types.GeneratorType)):
        for x in arg:
            for y in _flatten_ast_nodes(x):
                yield y
    else:
        raise TypeError(
            "_flatten_ast_nodes: unexpected %s" % (type(arg).__name__,))


def _iter_child_nodes_in_order(node):
    """
    Yield all direct child nodes of ``node``, that is, all fields that are nodes
    and all items of fields that are lists of nodes.

    ``_iter_child_nodes_in_order`` yields nodes in the same order that they
    appear in the source.

    ``ast.iter_child_nodes`` does the same thing, but not in source order.
    e.g. for ``Dict`` s, it yields all key nodes before all value nodes.
    """
    return _flatten_ast_nodes(_iter_child_nodes_in_order_internal_1(node))


def _iter_child_nodes_in_order_internal_1(node):
    if isinstance(node, str):
        # this happen for type comments which are not ast nodes but str
        # they do not have children. We yield nothing.
        yield []
        return
    if not isinstance(node, ast.AST):
        raise TypeError
    if isinstance(node, ast.Dict):
        assert node._fields == ("keys", "values")
        yield list(zip(node.keys, node.values))
    elif isinstance(node, (ast.FunctionDef, AsyncFunctionDef)):
        if sys.version_info < (3,12):
            assert node._fields == (
                "name",
                "args",
                "body",
                "decorator_list",
                "returns",
                "type_comment",
            ), node._fields
            res = (
                node.type_comment,
                node.decorator_list,
                node.args,
                node.returns,
                node.body,
            )
            yield res
        else:
            assert node._fields == (
                "name",
                "args",
                "body",
                "decorator_list",
                "returns",
                "type_comment",
                "type_params"
            ), node._fields
            res = (
                node.type_comment,
                node.decorator_list,
                node.args,
                node.returns,
                node.body,
                node.type_params
            )
            yield res


        # node.name is a string, not an AST node
    elif isinstance(node, ast.arguments):
        assert node._fields == ('posonlyargs', 'args', 'vararg', 'kwonlyargs',
                                'kw_defaults', 'kwarg', 'defaults'), node._fields
        args = node.posonlyargs + node.args
        defaults = node.defaults or ()
        num_no_default = len(args) - len(defaults)
        yield args[:num_no_default]
        yield list(zip(args[num_no_default:], defaults))
        # node.varags and node.kwarg are strings, not AST nodes.
    elif isinstance(node, ast.IfExp):
        assert node._fields == ('test', 'body', 'orelse')
        yield node.body, node.test, node.orelse
    elif isinstance(node, ast.Call):
        # call arguments order are lost by ast, re-order them
        yield node.func
        args = sorted([(k.value.lineno, k.value.col_offset, k) for k in node.keywords]+
                      [(k.lineno,k.col_offset, k) for k in node.args])
        yield [a[2] for a in args]
    elif isinstance(node, ast.ClassDef):
        if sys.version_info > (3, 12):
            assert node._fields == ('name', 'bases', 'keywords', 'body', 'decorator_list', 'type_params'), node._fields
            yield node.decorator_list, node.bases, node.body, node.type_params
        else:
            assert node._fields == ('name', 'bases', 'keywords', 'body', 'decorator_list'), node._fields
            yield node.decorator_list, node.bases, node.body
        # node.name is a string, not an AST node
    elif isinstance(node, ast.FormattedValue):
        assert node._fields == ('value', 'conversion', 'format_spec')
        yield node.value,
    elif isinstance(node, MatchAs):
        yield node.pattern
        yield node.name,
    elif isinstance(node, MatchMapping):
        for k, p in zip(node.keys, node.patterns):
            pass
            yield k, p
    else:
        # Default behavior.
        yield ast.iter_child_nodes(node)


def _walk_ast_nodes_in_order(node):
    """
    Recursively yield all child nodes of ``node``, in the same order that the
    node appears in the source.

    ``ast.walk`` does the same thing, but yields nodes in an arbitrary order.
    """
    # The implementation is basically the same as ``ast.walk``, but:
    #   1. Use a stack instead of a deque.  (I.e., depth-first search instead
    #      of breadth-first search.)
    #   2. Use _iter_child_nodes_in_order instead of ``ast.iter_child_nodes``.
    todo = [node]
    while todo:
        node = todo.pop()
        yield node
        todo.extend(reversed(list(_iter_child_nodes_in_order(node))))


def _flags_to_try(source:str, flags, auto_flags, mode):
    """
    Flags to try for ``auto_flags``.

    If ``auto_flags`` is False, then only yield ``flags``.
    If ``auto_flags`` is True, then yield ``flags`` and ``flags ^ print_function``.
    """
    flags = CompilerFlags(flags)
    if re.search(r"# *type:", source):
        flags = flags | CompilerFlags('type_comments')
    yield flags
    return


def _parse_ast_nodes(text:FileText, flags:CompilerFlags, auto_flags:bool, mode:str):
    """
    Parse a block of lines into an AST.

    Also annotate ``input_flags``, ``source_flags``, and ``flags`` on the
    resulting ast node.

    :type text:
      ``FileText``
    :type flags:
      ``CompilerFlags``
    :type auto_flags:
      ``bool``
    :param auto_flags:
      Whether to guess different flags if ``text`` can't be parsed with
      ``flags``.
    :param mode:
      Compilation mode: "exec", "single", or "eval".
    :rtype:
      ``ast.Module``
    """
    assert isinstance(text, FileText)
    filename = str(text.filename) if text.filename else "<unknown>"
    source = text.joined
    source = dedent(source)
    if not source.endswith("\n"):
        # Ensure that the last line ends with a newline (``ast`` barfs
        # otherwise).
        source += "\n"
    exp = None
    for flags in _flags_to_try(source, flags, auto_flags, mode):
        cflags = ast.PyCF_ONLY_AST | int(flags)
        try:
            result = compile(
                source, filename, mode, flags=cflags, dont_inherit=True)
        except SyntaxError as e:
            exp = e
            pass
        else:
            # Attach flags to the result.
            result.input_flags = flags
            result.source_flags = CompilerFlags.from_ast(result)
            result.flags = result.input_flags | result.source_flags
            result.text = text
            return result
    # None, would be unraisable and Mypy would complains below
    assert exp is not None
    raise exp



def _test_parse_string_literal(text:str, flags:CompilerFlags):
    r"""
    Attempt to parse ``text``.  If it parses cleanly to a single string
    literal, return its value.  Otherwise return ``None``.

      >>> _test_parse_string_literal(r'"foo\n" r"\nbar"', None)
      'foo\n\\nbar'

    """
    filetext = FileText(text)
    try:
        module_node = _parse_ast_nodes(filetext, flags, False, "eval")
    except SyntaxError:
        return None
    body = module_node.body
    if not _is_ast_str_or_byte(body):
        return None
    if sys.version_info < (3 ,9):
        return body.s
    else:
        return body.value


AstNodeContext = namedtuple("AstNodeContext", "parent field index")


def _annotate_ast_nodes(ast_node: ast.AST) -> AnnotatedAst:
    """
    Annotate AST with:
      - startpos and endpos
      - [disabled for now: context as `AstNodeContext` ]

    :type ast_node:
      ``ast.AST``
    :param ast_node:
      AST node returned by `_parse_ast_nodes`
    :return:
      ``None``
    """
    aast_node: AnnotatedAst = ast_node  # type: ignore
    text = aast_node.text
    flags = aast_node.flags
    startpos = text.startpos
    _annotate_ast_startpos(aast_node, None, startpos, text, flags)
    return aast_node


def _annotate_ast_startpos(
    ast_node: ast.AST, parent_ast_node, minpos: FilePos, text: FileText, flags
) -> bool:
    r"""
    Annotate ``ast_node``.  Set ``ast_node.startpos`` to the starting position
    of the node within ``text``.

    For "typical" nodes, i.e. those other than multiline strings, this is
    simply FilePos(ast_node.lineno, ast_node.col_offset+1), but taking
    ``text.startpos`` into account.

    For multiline string nodes, this function works by trying to parse all
    possible subranges of lines until finding the range that is syntactically
    valid and matches ``value``.  The candidate range is
    text[min_start_lineno:lineno+text.startpos.lineno+1].

    This function is unfortunately necessary because of a flaw in the output
    produced by the Python built-in parser.  For some crazy reason, the
    ``ast_node.lineno`` attribute represents something different for multiline
    string literals versus all other statements.  For multiline string literal
    nodes and statements that are just a string expression (or more generally,
    nodes where the first descendant leaf node is a multiline string literal),
    the compiler attaches the ending line number as the value of the ``lineno``
    attribute.  For all other than AST nodes, the compiler attaches the
    starting line number as the value of the ``lineno`` attribute.  This means
    e.g. the statement "'''foo\nbar'''" has a lineno value of 2, but the
    statement "x='''foo\nbar'''" has a lineno value of 1.

    :type ast_node:
      ``ast.AST``
    :type minpos:
      ``FilePos``
    :param minpos:
      Earliest position to check, in the number space of ``text``.
    :type text:
      ``FileText``
    :param text:
      Source text that was used to parse the AST, whose ``startpos`` should be
      used in interpreting ``ast_node.lineno`` (which always starts at 1 for
      the subset that was parsed).
    :type flags:
      ``CompilerFlags``
    :param flags:
      Compiler flags to use when re-compiling code.
    :return:
      ``True`` if this node is a multiline string literal or the first child is
      such a node (recursively); ``False`` otherwise.
    :raise ValueError:
      Could not find the starting line number.
    """
    assert isinstance(ast_node, (ast.AST, str, TypeIgnore)), ast_node
    aast_node: AnnotatedAst = cast(AnnotatedAst, ast_node)

    # First, traverse child nodes.  If the first child node (recursively) is a
    # multiline string, then we need to transfer its information to this node.
    # Walk all nodes/fields of the AST.  We implement this as a custom
    # depth-first search instead of using ast.walk() or ast.NodeVisitor
    # so that we can easily keep track of the preceding node's lineno.
    child_minpos: FilePos = minpos
    is_first_child: bool = True
    leftstr_node = None
    for child_node in _iter_child_nodes_in_order(aast_node):
        leftstr = _annotate_ast_startpos(
            child_node, aast_node, child_minpos, text, flags
        )
        if is_first_child and leftstr:
            leftstr_node = child_node
        if hasattr(child_node, 'lineno') and not isinstance(child_node, TypeIgnore):
            if child_node.startpos < child_minpos:
                raise AssertionError(
                    "Got out-of-order AST node(s):\n"
                    "  parent minpos=%s\n" % minpos
                    + "    node: %s\n" % ast.dump(aast_node)
                    + "      fields: %s\n" % (" ".join(aast_node._fields))
                    + "      children:\n"
                    + "".join(
                        "        %s %9s: %s\n"
                        % (
                            ("==>" if cn is child_node else "   "),
                            getattr(cn, "startpos", ""),
                            ast.dump(cn),
                        )
                        for cn in _iter_child_nodes_in_order(aast_node)
                    )
                    + "\n"
                    "This indicates a bug in pyflyby._\n"
                    "\n"
                    "pyflyby developer: Check if there's a bug or missing ast node handler in "
                    "pyflyby._parse._iter_child_nodes_in_order() - "
                    "probably the handler for ast.%s." % type(aast_node).__name__
                )
            child_minpos = child_node.startpos
        is_first_child = False

    # If the node has no lineno at all, then skip it.  This should only happen
    # for nodes we don't care about, e.g. ``ast.Module`` or ``ast.alias``.
    if not hasattr(aast_node, "lineno") or isinstance(aast_node, TypeIgnore):
        return False
    # If col_offset is set then the lineno should be correct also.
    if aast_node.col_offset >= 0:
        # In Python 3.8+, FunctionDef.lineno is the line with the def. To
        # account for decorators, we need the lineno of the first decorator
        if (
            isinstance(aast_node, (ast.FunctionDef, ast.ClassDef, AsyncFunctionDef))
            and aast_node.decorator_list
        ):
            delta = (
                aast_node.decorator_list[0].lineno - 1,
                # The col_offset doesn't include the @
                aast_node.decorator_list[0].col_offset - 1,
            )
        else:
            delta = (aast_node.lineno - 1, aast_node.col_offset)

        # Not a multiline string literal.  (I.e., it could be a non-string or
        # a single-line string.)
        # Easy.
        if sys.version_info < (3, 12):
            """There is an issue for f-strings at the begining of a file in 3.11 and
            before

            https://github.com/deshaw/pyflyby/issues/361,
            here we ensure a child node min pos, can't be before it's parent.
            """
            startpos = max(text.startpos + delta, minpos)
        else:
            startpos = text.startpos + delta

        # Special case for 'with' statements.  Consider the code:
        #    with X: pass
        #    ^0   ^5
        # Since 'Y's col_offset isn't the beginning of the line, the authors
        # of Python presumably changed 'X's col_offset to also not be the
        # beginning of the line.  If they had made the With ast node support
        # multiple clauses, they wouldn't have needed to do that, but then
        # that would introduce an API change in the AST.  So it's
        # understandable that they did that.
        # Since we use startpos for breaking lines, we need to set startpos to
        # the beginning of the line.
        # In Python 3, the col_offset for the with is 0 again.
        aast_node.startpos = startpos
        return False

    assert aast_node.col_offset == -1
    if leftstr_node:
        # This is an ast node where the leftmost deepest leaf is a
        # multiline string.  The bug that multiline strings have broken
        # lineno/col_offset infects ancestors up the tree.
        #
        # If the leftmost leaf is a multi-line string, then ``lineno``
        # contains the ending line number, and col_offset is -1:
        #   >>> ast.parse("""'''foo\nbar'''+blah""").body[0].lineno
        #   2
        # But if the leftmost leaf is not a multi-line string, then
        # ``lineno`` contains the starting line number:
        #   >>> ast.parse("""'''foobar'''+blah""").body[0].lineno
        #   1
        #   >>> ast.parse("""blah+'''foo\nbar'''+blah""").body[0].lineno
        #   1
        #
        # To fix that, we copy start_lineno and start_colno from the Str
        # node once we've corrected the values.
        assert not _is_ast_str_or_byte(aast_node)
        assert leftstr_node.lineno == aast_node.lineno
        assert leftstr_node.col_offset == -1
        aast_node.startpos = leftstr_node.startpos
        return True

    # a large chunk of what look like unreachable code has been removed from here
    # as the type annotation say many things were impossible (slices indexed by FilePos
    # instead of integers.
    raise ValueError("Couldn't find exact position of %s" % (ast.dump(ast_node)))


def _split_code_lines(ast_nodes, text):
    """
    Split the given ``ast_nodes`` and corresponding ``text`` by code/noncode
    statement.

    Yield tuples of (nodes, subtext).  ``nodes`` is a list of ``ast.AST`` nodes,
    length 0 or 1; ``subtext`` is a `FileText` sliced from ``text``.

    FileText(...))} for code lines and ``(None, FileText(...))`` for non-code
    lines (comments and blanks).

    :type ast_nodes:
      sequence of ``ast.AST`` nodes
    :type text:
      `FileText`
    """
    if not ast_nodes:
        yield ([], text)
        return
    assert text.startpos <= ast_nodes[0].startpos
    assert ast_nodes[-1].startpos < text.endpos
    if text.startpos != ast_nodes[0].startpos:
        # Starting noncode lines.
        yield ([], text[text.startpos:ast_nodes[0].startpos])
    end_sentinel = _DummyAst_Node()
    end_sentinel.startpos = text.endpos
    for node, next_node in zip(ast_nodes, ast_nodes[1:] + [end_sentinel]):
        startpos = node.startpos
        next_startpos = next_node.startpos
        assert startpos < next_startpos
        # We have the start position of this node.  Figure out the end
        # position, excluding noncode lines (standalone comments and blank
        # lines).
        if hasattr(node, 'endpos'):
            # We have an endpos for the node because this was a multi-line
            # string.  Start with the node endpos.
            endpos = node.endpos
            assert startpos < endpos <= next_startpos
            # enpos points to the character *after* the ending quote, so we
            # know that this is never at the beginning of the line.
            assert endpos.colno != 1
            # Advance past whitespace an inline comment, if any.  Do NOT
            # advance past other code that could be on the same line, nor past
            # blank lines and comments on subsequent lines.
            line = text[endpos : min(text.endpos, FilePos(endpos.lineno+1,1))]
            if _is_comment_or_blank(line):
                endpos = FilePos(endpos.lineno+1, 1)
        else:
            endpos = next_startpos
            assert endpos <= text.endpos
            # We don't have an endpos yet; what we do have is the next node's
            # startpos (or the position at the end of the text).  Start there
            # and work backward.
            if endpos.colno != 1:
                if endpos == text.endpos:
                    # There could be a comment on the last line and no
                    # trailing newline.
                    # TODO: do this in a more principled way.
                    if _is_comment_or_blank(text[endpos.lineno]):
                        assert startpos.lineno < endpos.lineno
                        if not text[endpos.lineno-1].endswith("\\"):
                            endpos = FilePos(endpos.lineno,1)
                else:
                    # We're not at end of file, yet the next node starts in
                    # the middle of the line.  This should only happen with if
                    # we're not looking at a comment.  [The first character in
                    # the line could still be "#" if we're inside a multiline
                    # string that's the last child of the parent node.
                    # Therefore we don't assert 'not
                    # _is_comment_or_blank(...)'.]
                    pass
            if endpos.colno == 1:
                while (endpos.lineno-1 > startpos.lineno and
                       _is_comment_or_blank(text[endpos.lineno-1]) and
                       (not text[endpos.lineno-2].endswith("\\") or
                        _is_comment_or_blank(text[endpos.lineno-2]))):
                    endpos = FilePos(endpos.lineno-1, 1)
        assert startpos < endpos <= next_startpos
        yield ([node], text[startpos:endpos])
        if endpos != next_startpos:
            yield ([], text[endpos:next_startpos])


def infer_compile_mode(arg:ast.AST) -> Literal['exec','eval','single']:
    """
    Infer the mode needed to compile ``arg``.

    :type arg:
      ``ast.AST``
    :rtype:
      ``str``
    """
    # Infer mode from ast object.
    if isinstance(arg, ast.Module):
        return "exec"
    elif isinstance(arg, ast.Expression):
        return "eval"
    elif isinstance(arg, ast.Interactive):
        return "single"
    else:
        raise TypeError(
            "Expected Module/Expression/Interactive ast node; got %s"
            % (type(arg).__name__))


class _DummyAst_Node:
    pass


class PythonStatement:
    r"""
    Representation of a top-level Python statement or consecutive
    comments/blank lines.

      >>> PythonStatement('print("x",\n file=None)\n')  #doctest: +SKIP
      PythonStatement('print("x",\n file=None)\n')

    Implemented as a wrapper around a `PythonBlock` containing at most one
    top-level AST node.
    """

    block: "PythonBlock"

    def __new__(cls, arg:Union[FileText, str], filename=None, startpos=None):

        if not isinstance(arg, (FileText, str)):
            raise TypeError("PythonStatement: unexpected %s" % type(arg).__name__)

        block = PythonBlock(arg, filename=filename, startpos=startpos)

        return cls.from_block(block)

    @classmethod
    def from_statement(cls, statement):
        assert isinstance(statement, cls), (statement, cls)
        return statement

    @classmethod
    def from_block(cls, block:PythonBlock) -> PythonStatement:
        """
        Return a statement from a PythonBlock

        This assume the PythonBlock is a single statement and check the comments
        to not start with newlines.
        """
        statements = block.statements
        if len(statements) != 1:
            raise ValueError(
                "Code contains %d statements instead of exactly 1: %r"
                % (len(statements), block)
            )
        (statement,) = statements
        assert isinstance(statement, cls)
        if statement.is_comment:
            assert not statement.text.joined.startswith("\n")
        return statement


    @classmethod
    def _construct_from_block(cls, block:PythonBlock):
        # Only to be used by PythonBlock.
        assert isinstance(block, PythonBlock), repr(block)
        self = object.__new__(cls)
        self.block = block
        if self.is_comment:
            assert not self.text.joined.startswith("\n"), self.text.joined
        return self

    @property
    def text(self) -> FileText:
        """
        :rtype:
          `FileText`
        """
        return self.block.text

    @property
    def filename(self) -> Optional[Filename]:
        """
        :rtype:
          `Filename`
        """
        return self.text.filename

    @property
    def startpos(self):
        """
        :rtype:
          `FilePos`
        """
        return self.text.startpos

    @property
    def flags(self):
        """
        :rtype:
          `CompilerFlags`
        """
        return self.block.flags

    @property
    def ast_node(self):
        """
        A single AST node representing this statement, or ``None`` if this
        object only represents comments/blanks.

        :rtype:
          ``ast.AST`` or ``NoneType``
        """
        ast_nodes = self.block.ast_node.body
        if len(ast_nodes) == 0:
            return None
        if len(ast_nodes) == 1:
            return ast_nodes[0]
        raise AssertionError("More than one AST node in block")

    @property
    def is_blank(self):
        return self.ast_node is None and self.text.joined.strip() == ''

    @property
    def is_comment(self):
        return self.ast_node is None and self.text.joined.strip() != ''

    @property
    def is_comment_or_blank(self):
        return self.is_comment or self.is_blank

    @property
    def is_comment_or_blank_or_string_literal(self):
        return (self.is_comment_or_blank
                or _ast_str_literal_value(self.ast_node) is not None)

    @property
    def is_import(self):
        return isinstance(self.ast_node, (ast.Import, ast.ImportFrom))

    @property
    def is_single_assign(self):
        n = self.ast_node
        return isinstance(n, ast.Assign) and len(n.targets) == 1

    def get_assignment_literal_value(self):
        """
        If the statement is an assignment, return the name and literal value.

          >>> PythonStatement('foo = {1: {2: 3}}').get_assignment_literal_value()
          ('foo', {1: {2: 3}})

        :return:
          (target, literal_value)
        """
        if not self.is_single_assign:
            raise ValueError(
                "Statement is not an assignment to a single name: %s" % self)
        n = self.ast_node
        target_name = n.targets[0].id
        literal_value = ast.literal_eval(n.value)
        return (target_name, literal_value)

    def __repr__(self):
        r = repr(self.block)
        assert r.startswith("PythonBlock(")
        r = "PythonStatement(" + r[12:]
        return r

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, PythonStatement):
            return NotImplemented
        return self.block == other.block

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if not isinstance(other, PythonStatement):
            return NotImplemented
        return self.block < other.block

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, PythonStatement):
            return NotImplemented
        return cmp(self.block, other.block)

    def __hash__(self):
        return hash(self.block)


class AnnotatedAst(ast.AST):
    text: FileText
    flags: str
    source_flags: CompilerFlags
    startpos: FilePos
    endpos: FilePos
    lieneno: int
    col_offset: int
    value: AnnotatedAst
    s: str


class AnnotatedModule(ast.Module, AnnotatedAst):
    source_flags: CompilerFlags


@total_ordering
class PythonBlock:
    r"""
    Representation of a sequence of consecutive top-level
    `PythonStatement` (s).

      >>> source_code = '# 1\nprint(2)\n# 3\n# 4\nprint(5)\nx=[6,\n 7]\n# 8\n'
      >>> codeblock = PythonBlock(source_code)
      >>> for stmt in PythonBlock(codeblock).statements:
      ...     print(stmt)
      PythonStatement('# 1\n')
      PythonStatement('print(2)\n', startpos=(2,1))
      PythonStatement('# 3\n# 4\n', startpos=(3,1))
      PythonStatement('print(5)\n', startpos=(5,1))
      PythonStatement('x=[6,\n 7]\n', startpos=(6,1))
      PythonStatement('# 8\n', startpos=(8,1))

    A ``PythonBlock`` has a ``flags`` attribute that gives the compiler_flags
    associated with the __future__ features using which the code should be
    parsed.

    """

    text: FileText
    _auto_flags: bool
    _input_flags: Union[int,CompilerFlags]

    def __new__(cls, arg:Any, filename=None, startpos=None, flags=None,
                auto_flags=None):
        if isinstance(arg, PythonStatement):
            arg = arg.block
            # Fall through
        if isinstance(arg, cls):
            if filename is startpos is flags is None:
                return arg
            flags = CompilerFlags(flags, arg.flags)
            arg = arg.text
            # Fall through
        if isinstance(arg, (FileText, Filename, str)):
            return cls.from_text(
                arg, filename=filename, startpos=startpos,
                flags=flags, auto_flags=auto_flags)
        raise TypeError("%r: unexpected %r"
                        % (cls.__name__, type(arg).__name__,))

    @classmethod
    def from_filename(cls, filename):
        return cls.from_text(Filename(filename))

    @classmethod
    def from_text(
        cls, text, filename=None, startpos=None, flags=None, auto_flags: bool = False
    ):
        """
        :type text:
          `FileText` or convertible
        :type filename:
          ``Filename``
        :param filename:
          Filename, if not already given by ``text``.
        :type startpos:
          ``FilePos``
        :param startpos:
          Starting position, if not already given by ``text``.
        :type flags:
          ``CompilerFlags``
        :param flags:
          Input compiler flags.
        :param auto_flags:
          Whether to try other flags if ``flags`` fails.
        :rtype:
          `PythonBlock`
        """
        if isinstance(filename, str):
            filename = Filename(filename)
        assert isinstance(filename, (Filename, NoneType)), filename
        text = FileText(text, filename=filename, startpos=startpos)
        self = object.__new__(cls)
        self.text = text
        self._input_flags = CompilerFlags(flags)
        self._auto_flags = auto_flags
        return self

    @classmethod
    def __construct_from_annotated_ast(cls, annotated_ast_nodes, text:FileText, flags):
        # Constructor for internal use by _split_by_statement() or
        # concatenate().
        ast_node = AnnotatedModule(annotated_ast_nodes, type_ignores=[])
        ast_node.text = text
        ast_node.flags = flags
        if not hasattr(ast_node, "source_flags"):
            ast_node.source_flags = CompilerFlags.from_ast(annotated_ast_nodes)
        self = object.__new__(cls)
        self._ast_node_or_parse_exception = ast_node
        self.ast_node                     = ast_node
        self.annotated_ast_node           = ast_node
        self.text                         = text
        self.flags                        = flags
        self._input_flags                 = flags
        self._auto_flags                  = False
        return self

    @classmethod
    def concatenate(cls, blocks, assume_contiguous=_sentinel):
        """
        Concatenate a bunch of blocks into one block.

        :type blocks:
          sequence of `PythonBlock` s and/or `PythonStatement` s
        :param assume_contiguous:
          Deprecated, always True
          Whether to assume, without checking, that the input blocks were
          originally all contiguous.  This must be set to True to indicate the
          caller understands the assumption; False is not implemented.
        """
        if assume_contiguous is not _sentinel:
            warnings.warn('`assume_continuous` is deprecated and considered always `True`')
            assume_contiguous = True
        if not assume_contiguous:
            raise NotImplementedError
        blocks2 = []
        for b in blocks:
            if isinstance(b, PythonStatement):
                b = b.block
            if not isinstance(b, PythonBlock):
                b = PythonBlock(b)
            blocks2.append(b)
        blocks = blocks2
        if len(blocks) == 1:
            return blocks[0]
        assert blocks
        text = FileText.concatenate([b.text for b in blocks])
        # The contiguous assumption is important here because ``ast_node``
        # contains line information that would otherwise be wrong.
        ast_nodes = [n for b in blocks for n in b.annotated_ast_node.body]
        flags = blocks[0].flags
        return cls.__construct_from_annotated_ast(ast_nodes, text, flags)

    @property
    def filename(self):
        return self.text.filename

    @property
    def startpos(self):
        return self.text.startpos

    @property
    def endpos(self):
        return self.text.endpos

    @cached_property
    def _ast_node_or_parse_exception(self):
        """
        Attempt to parse this block of code into an abstract syntax tree.
        Cached (including exception case).

        :return:
          Either ast_node or exception.
        """
        # This attribute may also be set by __construct_from_annotated_ast(),
        # in which case this code does not run.
        try:
            return _parse_ast_nodes(
                self.text, self._input_flags, self._auto_flags, "exec")
        except Exception as e:
            # Add the filename to the exception message to be nicer.
            if self.text.filename:
                try:
                    e = type(e)("While parsing %s: %s" % (self.text.filename, e))
                except TypeError:
                    # Exception takes more than one argument
                    pass
            # Cache the exception to avoid re-attempting while debugging.
            return e

    @cached_property
    def parsable(self):
        """
        Whether the contents of this ``PythonBlock`` are parsable as Python
        code, using the given flags.

        :rtype:
          ``bool``
        """
        return isinstance(self._ast_node_or_parse_exception, ast.AST)

    @cached_property
    def parsable_as_expression(self):
        """
        Whether the contents of this ``PythonBlock`` are parsable as a single
        Python expression, using the given flags.

        :rtype:
          ``bool``
        """
        return self.parsable and self.expression_ast_node is not None

    @cached_property
    def ast_node(self):
        """
        Parse this block of code into an abstract syntax tree.

        The returned object type is the kind of AST as returned by the
        ``compile`` built-in (rather than as returned by the older, deprecated
        ``compiler`` module).  The code is parsed using mode="exec".

        The result is a ``ast.Module`` node, even if this block represents only
        a subset of the entire file.

        :rtype:
          ``ast.Module``
        """
        r = self._ast_node_or_parse_exception
        if isinstance(r, ast.AST):
            return r
        else:
            raise r

    @cached_property
    def annotated_ast_node(self) -> AnnotatedAst:
        """
        Return ``self.ast_node``, annotated in place with positions.

        All nodes are annotated with ``startpos``.
        All top-level nodes are annotated with ``endpos``.

        :rtype:
          ``ast.Module``
        """
        result = self.ast_node
        # ! result is mutated and returned
        return _annotate_ast_nodes(result)

    @cached_property
    def expression_ast_node(self) -> Optional[ast.Expression]:
        """
        Return an ``ast.Expression`` if ``self.ast_node`` can be converted into
        one.  I.e., return parse(self.text, mode="eval"), if possible.

        Otherwise, return ``None``.

        :rtype:
          ``ast.Expression``
        """
        node = self.ast_node
        if len(node.body) == 1 and isinstance(node.body[0], ast.Expr):
            return ast.Expression(node.body[0].value)
        else:
            return None

    def parse(self, mode: Optional[str] = None) -> Union[ast.Expression, ast.Module]:
        """
        Parse the source text into an AST.

        :param mode:
          Compilation mode: "exec", "single", or "eval".  "exec", "single",
          and "eval" work as the built-in ``compile`` function do.  If ``None``,
          then default to "eval" if the input is a string with a single
          expression, else "exec".
        :rtype:
          ``ast.AST``
        """
        if mode == "exec":
            assert isinstance(self.ast_node, ast.Module)
            return self.ast_node
        elif mode == "eval":
            if self.expression_ast_node:
                assert isinstance(self.ast_node, ast.Expression)
                return self.expression_ast_node
            else:
                raise SyntaxError
        elif mode is None:
            if self.expression_ast_node:
                return self.expression_ast_node
            else:
                assert isinstance(self.ast_node, ast.Module)
                return self.ast_node
        elif mode == "exec":
            raise NotImplementedError
        else:
            raise ValueError("parse(): invalid mode=%r" % (mode,))

    def compile(self, mode: Optional[str] = None):
        """
        Parse into AST and compile AST into code.

        :rtype:
          ``CodeType``
        """
        ast_node = self.parse(mode=mode)
        c_mode = infer_compile_mode(ast_node)
        filename = str(self.filename or "<unknown>")
        return compile(ast_node, filename, c_mode)

    @cached_property
    def statements(self) -> Tuple[PythonStatement, ...]:
        r"""
        Partition of this ``PythonBlock`` into individual ``PythonStatement`` s.
        Each one contains at most 1 top-level ast node.  A ``PythonStatement``
        can contain no ast node to represent comments.

          >>> code = "# multiline\n# comment\n'''multiline\nstring'''\nblah\n"
          >>> print(PythonBlock(code).statements) # doctest:+NORMALIZE_WHITESPACE
          (PythonStatement('# multiline\n# comment\n'),
           PythonStatement("'''multiline\nstring'''\n", startpos=(3,1)),
           PythonStatement('blah\n', startpos=(5,1)))

        :rtype:
          ``tuple`` of `PythonStatement` s
        """
        node = self.annotated_ast_node
        nodes_subtexts = list(_split_code_lines(node.body, self.text))  # type: ignore
        cls = type(self)
        statement_blocks: List[PythonBlock] = [
            cls.__construct_from_annotated_ast(subnodes, subtext, self.flags)
            for subnodes, subtext in nodes_subtexts]

        no_newline_blocks = []
        for block in statement_blocks:
            # The ast parsing make "comments" start at the ends of the previous node,
            # so might including starting with blank lines. We never want blocks to
            # start with new liens or that messes up the formatting code that insert/count new lines.
            while block.text.joined.startswith("\n") and block.text.joined != "\n":
                first, *other = block.text.lines
                assert not first.endswith('\n')
                no_newline_blocks.append(
                    PythonBlock(
                        first+'\n',
                        filename=block.filename,
                        startpos=block.startpos,
                        flags=block.flags,
                    )
                )
                # assert block.startpos == (0,0), (block.startpos, block.text.joined)
                # just use lines 1: here and decrease startpos ?
                block = PythonBlock(
                    "\n".join(other),
                    filename=block.filename,
                    startpos=block.startpos,
                    flags=block.flags,
                )
            no_newline_blocks.append(block)

        # Convert to statements.
        statements = []
        for b in no_newline_blocks:
            assert isinstance(b, PythonBlock)
            statement = PythonStatement._construct_from_block(b)
            statements.append(statement)
        return tuple(statements)

    @cached_property
    def source_flags(self):
        """
        If the AST contains __future__ imports, then the compiler_flags
        associated with them.  Otherwise, 0.

        The difference between ``source_flags`` and ``flags`` is that ``flags``
        may be set by the caller (e.g. based on an earlier __future__ import)
        and include automatically guessed flags, whereas ``source_flags`` is
        only nonzero if this code itself contains __future__ imports.

        :rtype:
          `CompilerFlags`
        """
        return self.ast_node.source_flags

    @cached_property
    def flags(self):
        """
        The compiler flags for this code block, including both the input flags
        (possibly automatically guessed), and the flags from "__future__"
        imports in the source code text.

        :rtype:
          `CompilerFlags`
        """
        return self.ast_node.flags

    def groupby(self, predicate):
        """
        Partition this block of code into smaller blocks of code which
        consecutively have the same ``predicate``.

        :param predicate:
          Function that takes a `PythonStatement` and returns a value.
        :return:
          Generator that yields (group, `PythonBlock` s).
        """
        cls = type(self)
        for pred, stmts in groupby(self.statements, predicate):
            blocks = [s.block for s in stmts]
            yield pred, cls.concatenate(blocks)

    def string_literals(self):
        r"""
        Yield all string literals anywhere in this block.

        The string literals have ``startpos`` attributes attached.

          >>> block = PythonBlock("'a' + ('b' + \n'c')")
          >>> [(f.s, f.startpos) for f in block.string_literals()]
          [('a', FilePos(1,1)), ('b', FilePos(1,8)), ('c', FilePos(2,1))]

        :return:
          Iterable of ``ast.Str``  or ``ast.Bytes`` nodes
        """
        for node in _walk_ast_nodes_in_order(self.annotated_ast_node):
            if _is_ast_str_or_byte(node):
                assert hasattr(node, 'startpos')
                yield node

    def _get_docstring_nodes(self):
        """
        Yield docstring AST nodes.

        We consider the following to be docstrings::

          - First literal string of function definitions, class definitions,
            and modules (the python standard)
          - Literal strings after assignments, per Epydoc

        :rtype:
          Generator of ``ast.Str`` nodes
        """
        # This is similar to ``ast.get_docstring``, but:
        #   - This function is recursive
        #   - This function yields the node object, rather than the string
        #   - This function yields multiple docstrings (even per ast node)
        #   - This function doesn't raise TypeError on other AST types
        #   - This function doesn't cleandoc
        docstring_containers = (ast.FunctionDef, ast.ClassDef, ast.Module, AsyncFunctionDef)
        for node in _walk_ast_nodes_in_order(self.annotated_ast_node):
            if not isinstance(node, docstring_containers):
                continue
            if not node.body:
                continue
            # If the first body item is a literal string, then yield the node.
            if (isinstance(node.body[0], ast.Expr) and
                _is_ast_str(node.body[0].value)):
                yield node.body[0].value
            for i in range(1, len(node.body)-1):
                # If a body item is an assignment and the next one is a
                # literal string, then yield the node for the literal string.
                n1, n2 = node.body[i], node.body[i+1]
                if (isinstance(n1, ast.Assign) and
                    isinstance(n2, ast.Expr) and
                    _is_ast_str(n2.value)):
                    yield n2.value

    def get_doctests(self):
        r"""
        Return doctests in this code.

          >>> PythonBlock("x\n'''\n >>> foo(bar\n ...     + baz)\n'''\n").get_doctests()
          [PythonBlock('foo(bar\n    + baz)\n', startpos=(3,2))]

        :rtype:
          ``list`` of `PythonStatement` s
        """
        parser = IgnoreOptionsDocTestParser()
        doctest_blocks = []
        filename = self.filename
        flags = self.flags
        for ast_node in self._get_docstring_nodes():
            try:
                if sys.version_info >= (3, 10):
                    examples = parser.get_examples(ast_node.value)
                else:
                    examples = parser.get_examples(ast_node.s)
            except Exception:
                if sys.version_info >= (3, 10):
                    blob = ast_node.s
                else:
                    blob = ast_node.value
                if len(blob) > 60:
                    blob = blob[:60] + '...'
                # TODO: let caller decide how to handle
                logger.warning("Can't parse docstring; ignoring: %r", blob)
                continue
            for example in examples:
                lineno = ast_node.startpos.lineno + example.lineno
                colno = ast_node.startpos.colno + example.indent # dubious
                text = FileText(example.source, filename=filename,
                                startpos=(lineno,colno))
                try:
                    block = PythonBlock(text, flags=flags)
                    block.ast_node # make sure we can parse
                except Exception:
                    blob = text.joined
                    if len(blob) > 60:
                        blob = blob[:60] + '...'
                    logger.warning("Can't parse doctest; ignoring: %r", blob)
                    continue
                doctest_blocks.append(block)
        return doctest_blocks

    def __repr__(self):
        r = "%s(%r" % (type(self).__name__, self.text.joined)
        if self.filename:
            r += ", filename=%r" % (str(self.filename),)
        if self.startpos != FilePos():
            r += ", startpos=%s" % (self.startpos,)
        if self.flags != self.source_flags:
            r += ", flags=%s" % (self.flags,)
        r += ")"
        return r

    def __str__(self):
        return str(self.text)

    def __text__(self):
        return self.text

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, PythonBlock):
            return NotImplemented
        return self.text == other.text and self.flags == other.flags

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if not isinstance(other, PythonBlock):
            return NotImplemented
        return (self.text, self.flags) < (other.text, other.flags)

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, PythonBlock):
            return NotImplemented
        return cmp(self.text, other.text) or cmp(self.flags, other.flags)

    def __hash__(self):
        h = hash((self.text, self.flags))
        self.__hash__ = lambda: h
        return h

class IgnoreOptionsDocTestParser(DocTestParser):
    def _find_options(self, source, name, lineno):
        # Ignore doctest options. We don't use them, and we don't want to
        # error on unknown options, which is what the default DocTestParser
        # does.
        return {}



# pyflyby/test_file.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/



import pytest

from   pyflyby._file            import FilePos, FileText, Filename
from   pyflyby._util            import CwdCtx

def test_Filename_1():
    f = Filename('/etc/passwd')
    assert str(f) == '/etc/passwd'


def test_Filename_eqne_1():
    assert      Filename('/foo/bar') == Filename('/foo/bar')
    assert not (Filename('/foo/bar') != Filename('/foo/bar'))
    assert      Filename('/foo/bar') != Filename('/foo/BAR')
    assert not (Filename('/foo/bar') == Filename('/foo/BAR'))


def test_Filename_eqne_other_1():
    assert      Filename("/foo") != "/foo"
    assert not (Filename("/foo") == "/foo")
    assert      Filename("/foo") != object()
    assert not (Filename("/foo") == object())


def test_Filename_abspath_1():
    with CwdCtx("/dev"):
        f = Filename("foo")
    assert f == Filename("/dev/foo")


def test_Filename_normpath_1():
    with CwdCtx("/dev"):
        f = Filename("../a/b/../c")
    assert f == Filename("/a/c")


def test_Filename_dir_1():
    f = Filename('/Foo.foo.f/Bar.bar.b/Quux.quux.q')
    assert f.dir == Filename('/Foo.foo.f/Bar.bar.b')


def test_Filename_base_1():
    f = Filename('/Foo.foo.f/Bar.bar.b/Quux.quux.q')
    assert f.base == 'Quux.quux.q'


def test_Filename_ext_1():
    f = Filename('/Foo.foo.f/Bar.bar.b/Quux.quux.q')
    assert f.ext == '.q'


def test_Filename_isfile_1():
    f = Filename('/etc/passwd')
    assert f.isfile
    assert not f.dir.isfile


def test_Filename_isdir_1():
    f = Filename('/etc/passwd')
    assert not f.isdir
    assert f.dir.isdir


def test_Filename_ancestors_1():
    fn = Filename("/a.aa/b.bb/c.cc")
    result = fn.ancestors
    expected = (Filename("/a.aa/b.bb/c.cc"),
                Filename("/a.aa/b.bb"),
                Filename("/a.aa"),
                Filename("/"))
    assert result == expected


def test_FilePos_from_intint_1():
    pos = FilePos(55,66)
    assert pos.lineno == 55
    assert pos.colno  == 66


def test_FilePos_from_FilePos_1():
    pos = FilePos(55,66)
    assert FilePos(pos) is pos


def test_FilePos_from_tuple_intint_1():
    pos = FilePos((66,77))
    assert pos.lineno == 66
    assert pos.colno  == 77


def test_FilePos_from_None_1():
    pos = FilePos(None)
    assert pos.lineno == 1
    assert pos.colno  == 1


def test_FilePos_from_empty_1():
    pos = FilePos()
    assert pos.lineno == 1
    assert pos.colno  == 1


def test_FilePos_eqne_lineno_1():
    p1a = FilePos(55,66)
    p1b = FilePos(55,66)
    p2  = FilePos(55,1)
    assert     (p1a == p1b)
    assert not (p1a != p1b)
    assert     (p1a != p2 )
    assert not (p1a == p2 )


def test_FilePos_eqne_colno_1():
    p1a = FilePos(55,66)
    p1b = FilePos(55,66)
    p2  = FilePos(1,66)
    assert     (p1a == p1b)
    assert not (p1a != p1b)
    assert     (p1a != p2 )
    assert not (p1a == p2 )


def test_FilePos_bad_other_1():
    with pytest.raises(TypeError):
        FilePos(object())


def test_FilePos_bad_too_few_args_1():
    with pytest.raises(TypeError):
        FilePos(5)
    with pytest.raises(TypeError):
        FilePos((5,))


def test_FilePos_bad_too_many_args_1():
    with pytest.raises(TypeError):
        FilePos(5,6,7)
    with pytest.raises(TypeError):
        FilePos((5,6,7))


def test_FilePos_bad_type_1():
    with pytest.raises(TypeError):
        FilePos("5","6")
    with pytest.raises(TypeError):
        FilePos(5.0, 6.0)


def test_FileText_from_str_1():
    text = FileText("a\nb\nc\nd")
    assert text.joined == "a\nb\nc\nd"
    assert text.lines == ("a", "b", "c", "d")


def test_FileText_from_str_trailing_newline_1():
    text = FileText("a\nb\nc\nd\n")
    assert text.joined == "a\nb\nc\nd\n"
    assert text.lines == ("a", "b", "c", "d", "")


def test_FileText_from_str_one_1():
    text = FileText("a")
    assert text.joined == "a"
    assert text.lines == ("a",)


def test_FileText_from_str_one_trailing_newline_1():
    text = FileText("a\n")
    assert text.joined == "a\n"
    assert text.lines == ("a", "")


def test_FileText_idempotent_1():
    text = FileText("a\nb\nc\nd")
    assert FileText(text) is text


def test_FileText_attr_defaults_1():
    text = FileText("aabb\n")
    assert text.joined   == "aabb\n"
    assert text.filename == None
    assert text.startpos.lineno   == 1
    assert text.startpos.colno    == 1


def test_FileText_attrs_1():
    text = FileText("aabb\n", filename="/foo", startpos=(100,5))
    assert text.joined          == "aabb\n"
    assert text.filename        == Filename("/foo")
    assert text.startpos.lineno == 100
    assert text.startpos.colno  == 5


def test_FileText_attrs_from_instance_1():
    text = FileText(FileText("aabb\n"), filename="/foo", startpos=(100,5))
    assert text.joined          == "aabb\n"
    assert text.filename        == Filename("/foo")
    assert text.startpos.lineno == 100
    assert text.startpos.colno  == 5


def test_FileText_endpos_1():
    text = FileText("foo\nbar\n")
    assert text.endpos == FilePos(3,1)


def test_FileText_endpos_trailing_partial_line_1():
    text = FileText("foo\nbar")
    assert text.endpos == FilePos(2,4)


def test_FileText_endpos_offset_1():
    text = FileText("foo\nbar\n", startpos=(101,55))
    assert text.endpos == FilePos(103,1)


def test_FileText_empty_1():
    text = FileText("", startpos=(5,5))
    assert text.lines == ("",)
    assert text.joined == ""
    assert text.startpos == text.endpos == FilePos(5,5)


def test_FileText_one_full_line_offset_1():
    text = FileText("foo\n", startpos=(101,55))
    assert text.endpos == FilePos(102,1)


def test_FileText_one_partial_line_offset_1():
    text = FileText("foo", startpos=(101,55))
    assert text.endpos == FilePos(101,58)


def test_FileText_getitem_1():
    assert FileText("a\nb\nc\nd")[2] == 'b'


def test_FileText_slice_1():
    assert FileText("a\nb\nc\nd")[2:4] == FileText('b\nc\n', startpos=(2,1))


def test_FileText_slice_offset_1():
    text = FileText("a\nb\nc\nd", startpos=(101,5))
    assert text[102:104] == FileText('b\nc\n', startpos=(102,1))


def test_FileText_slice_col_1():
    text = FileText("one\ntwo4567\nthree6789\nfour\n", startpos=(101,55))
    result = text[ (102,3) : (103,8) ]
    expected = FileText("o4567\nthree67", startpos=(102,3))
    assert result == expected


def test_FileText_slice_starting_col_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    result = text[ (102,103) : (103,8) ]
    expected = FileText("o4567\nthree67", startpos=(102,103))
    assert result == expected


def test_FileText_slice_col_eol_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    result = text[ (102,103) : (104,1) ]
    expected = FileText("o4567\nthree6789\n", startpos=(102,103))
    assert result == expected


def test_FileText_slice_col_eof_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    result = text[ (102,103) : (105,1) ]
    expected = FileText("o4567\nthree6789\nfour\n", startpos=(102,103))
    assert result == expected


def test_FileText_slice_idempotent_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    result = text[ (102,101) : (105,1) ]
    assert result is text


def test_FileText_slice_idempotent_2():
    text = FileText("two4567\nthree6789\nfour", startpos=(102,101))
    result = text[ (102,101) : (104,5) ]
    assert result is text


def test_FileText_slice_col_almost_eof_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    result = text[ (102,103) : (104,5) ]
    expected = FileText("o4567\nthree6789\nfour", startpos=(102,103))
    assert result == expected


def test_FileText_slice_col_out_of_range_start_lineno_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    with pytest.raises(IndexError):
        text[ (101,1) : (103,1) ]


def test_FileText_slice_col_out_of_range_end_lineno_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    with pytest.raises(IndexError):
        text[ (102,103) : (106,1) ]


def test_FileText_slice_col_out_of_range_start_colno_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    with pytest.raises(IndexError):
        text[ (102,1) : (103,1) ]


def test_FileText_slice_col_out_of_range_end_colno_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    with pytest.raises(IndexError):
        text[ (102,103) : (105,2) ]


def test_FileText_slice_col_out_of_range_end_colno_2():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    with pytest.raises(IndexError):
        text[ (102,103) : (104,6) ]


def test_FileText_slice_empty_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    result = text[ (102,103) : (102,103) ]
    expected = FileText("", startpos=(102,103))
    assert result == expected


def test_FileText_slice_out_of_range_empty_1():
    text = FileText("two4567\nthree6789\nfour\n", startpos=(102,101))
    with pytest.raises(IndexError):
        text[ (102,200) : (102,200) ]


def test_FileText_getitem_out_of_range_1():
    text = FileText("a\nb\nc\nd")
    with pytest.raises(IndexError):
        text[0]


def test_FileText_eqne_1():
    text1a = FileText("hello\n")
    text1b = FileText("hello\n")
    text2  = FileText("hello\nhello\n")
    assert     (text1a == text1b)
    assert not (text1a != text1b)
    assert     (text1a != text2 )
    assert not (text1a == text2 )


def test_FileText_eqne_lineno_1():
    text1a = FileText("hello\n", startpos=(100,1))
    text1b = FileText("hello\n", startpos=(100,1))
    text2  = FileText("hello\n", startpos=(101,1))
    assert     (text1a == text1b)
    assert not (text1a != text1b)
    assert     (text1a != text2 )
    assert not (text1a == text2 )


def test_FileText_eqne_filename_1():
    text1a = FileText("hello\n", filename='/foo')
    text1b = FileText("hello\n", filename='/foo')
    text2  = FileText("hello\n", filename='/bar')
    assert     (text1a == text1b)
    assert not (text1a != text1b)
    assert     (text1a != text2 )
    assert not (text1a == text2 )


def test_FileText_ne_other_1():
    text = FileText("hello\n")
    assert     (text != object())
    assert not (text == object())
    assert     (text != "hello\n")
    assert not (text == "hello\n")



from pyflyby._parse import PythonBlock
from pyflyby._imports2s import fix_unused_and_missing_imports
from pyflyby._import_sorting import sort_imports
from textwrap import dedent
from lib.python.pyflyby._importstmt import ImportFormatParams
import pytest

code1 = dedent("""
        import external
        import os
        import sympy
        import numpy
        import json
        import simplejson

        from pkg2 import same, line
        from pkg1.mod1 import foo
        from pkg1.mod2 import bar
        from pkg2 import baz
        import yy

        from pkg1.mod1 import foo2
        from pkg1.mod3 import quux
        from pkg2 import baar
        import zz
    """).strip()

# The logic requested in issue 13 was to keep blocks, but order
# lexicographically. we do seem to be splitting stdlib from installed packages,
# but not adding a blank line and then sort lexicographically if an package has
# several submodules imported we put it in a block, otherwise we keep
# everything together.

# note that sorting does not concatenate same imports together, but cannonicalize
# in tidy import will.

expected1 = dedent("""
        import external
        import json
        import numpy
        import os

        from pkg1.mod1                import foo
        from pkg1.mod1                import foo2
        from pkg1.mod2                import bar
        from pkg1.mod3                import quux

        from pkg2                     import same, line
        from pkg2                     import baz
        from pkg2                     import baar

        import simplejson
        import sympy
        import yy
        import zz
    """).strip()+'\n\n'


code2 = dedent(
    """
    '''module docstring'''
    import os

    pass
    """
).strip()

expected2 = code2


code3 = dedent("""
    import os

    if True:
        "ok"
    """).strip()

expected3 = code3

# stable should not change
stable_1 = dedent(
    """
        #!/usr/local/bin/python3

        from   deshaw.abc               import DAYS
        from   deshaw.py                import tuple

        print(DAYS.days(20240101, 20241231))
        print(tuple("hello"))
    """
)


code4 = dedent(
    """
from __future__ import print_function

from pyflyby._cmdline         import filename_args, hfmt, parse_args
from pyflyby._importclns      import ImportSet
from pyflyby._importdb        import ImportDB

import re
import sys

def main():
    def addopts(parser):
        pass"""
).strip()

expected4 = code4


@pytest.mark.parametrize(
    "code, expected",
    [(code1, expected1), (code2, expected2), (code3, expected3), (code4, expected4)],
)
def test_sort_1(code, expected):

    assert str(sort_imports(PythonBlock(code))) == expected
    # expected is stable
    assert str(sort_imports(PythonBlock(expected))) == expected


@pytest.mark.parametrize("code", [stable_1])
def test_stable(code):
    params = ImportFormatParams(
        align_imports=(32,),
        from_spaces=3,
        separate_from_imports=False,
        max_line_length=None,
        use_black=False,
        align_future=False,
        hanging_indent="never",
    )
    add_missing = False
    remove_unused = False
    add_mandatory = False

    result = fix_unused_and_missing_imports(
        PythonBlock(code),
        params=params,
        add_missing=add_missing,
        remove_unused=remove_unused,
        add_mandatory=add_mandatory,
    )

    assert str(result) == str(stable_1)



# pyflyby/_py.py
# Copyright (C) 2014, 2015, 2018, 2019 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT

r"""
The `py` program (part of the pyflyby project) is a command-line multitool for
running python code, with heuristic intention guessing, automatic importing,
and debugging support.

Invocation summary
==================

.. code::

  py [--file]   filename.py arg1 arg2   Execute a file
  py [--eval]  'function(arg1, arg2)'   Evaluate an expression/statement
  py [--apply]  function arg1 arg2      Call function(arg1, arg2)
  py [--module] modname arg1 arg2       Run a module

  py  --map     function arg1 arg2      Call function(arg1); function(arg2)

  py  -i       'function(arg1, arg2)'   Run file/code/etc, then run IPython
  py  --debug  'function(arg1, arg2)'   Debug file/code/etc
  py  --debug   PID                     Attach debugger to PID

  py            function?               Get help for a function or module
  py            function??              Get source of a function or module

  py                                    Start IPython with autoimporter
  py nb                                 Start IPython Notebook with autoimporter


  py [--add-deprecated-builtins]        Inject "breakpoint", "debug_exception",
                                        "debug_statement", "waitpoint" into
                                        builtins. This is deprecated, and
                                        present for backward compatibilty
                                        but will be removed in the future.

Features
========

  * Heuristic action mode guessing: If none of --file, --eval, --apply,
    --module, or --map is specified, then guess what to do, choosing one of
    these actions:

      * Execute (run) a file
      * Evaluate concatenated arguments
      * Run a module
      * Call (apply) a function
      * Evaluate first argument

  * Automatic importing: All action modes (except run_module) automatically
    import as needed.

  * Heuristic argument evaluation: By default, `py --eval`, `py --apply`, and
    `py --map` guess whether the arguments should be interpreted as
    expressions or literal strings. A "--" by itself will designate subsequent
    args as strings.  A "-" by itself will be replaced by the contents of
    stdin as a string.

  * Merged eval/exec: Code is eval()ed or exec()ed as appropriate.

  * Result printing: By default, results are pretty-printed if not None.

  * Heuristic flags: "print" can be used as a function or a statement.

  * Matplotlib/pylab integration: show() is called if appropriate to block on
    plots.

  * Enter debugger upon unhandled exception.  (This functionality is enabled
    by default when stdout is a tty.  Use --postmortem=no to never use the
    postmortem debugger.  Use --postmortem=yes enable even if stdout is not a
    tty.  If the postmortem debugger is enabled but /dev/tty is not available,
    then if an exception occurs, py will email instructions for attaching a
    debugger.)

  * Control-\\ (SIGQUIT) enters debugger while running (and allows continuing).

  * New builtin functions such as "debugger()".

Warning
=======
`py` is intended as an interactive tool.  When writing shell aliases for
interactive use, the `--safe` option can be useful.  When writing scripts,
it's better to avoid all heuristic guessing; use regular `python -c ...`, or
better yet, a full-fledged python program (and run tidy-imports).


Options
=======

.. code::

    Global options valid before code argument:

      --args=string    Interpret all arguments as literal strings.
                       (The "--" argument also specifies remaining arguments to be
                       literal strings.)
      --args=eval      Evaluate all arguments as expressions.
      --args=auto      (Default) Heuristically guess whether to evaluate arguments
                       as literal strings or expressions.
      --output=silent  Don't print the result of evaluation.
      --output=str     Print str(result).
      --output=repr    Print repr(result).
      --output=pprint  Print pprint.pformat(result).
      --output=repr-if-not-none
                       Print repr(result), but only if result is not None.
      --output=pprint-if-not-none
                       Print pprint.pformat(result), but only if result is not None.
      --output=interactive
                       (Default) Print result.__interactive_display__() if defined,
                       else pprint if result is not None.
      --output=exit    Raise SystemExit(result).
      --safe           Equivalent to --args=strings and PYFLYBY_PATH=EMPTY.
      --quiet, --q     Log only error messages to stderr; omit info and warnings.
      --interactive, --i
                       Run an IPython shell after completion
      --debug, --d     Run the target code file etc under the debugger.  If a PID is
                       given, then instead attach a debugger to the target PID.
      --verbose        Turn on verbose messages from pyflyby.

    Pseudo-actions valid before, after, or without code argument:

      --version        Print pyflyby version or version of a module.
      --help, --h, --? Print this help or help for a function or module.
      --source, --??   Print source code for a function or module.


Examples
========

  Start IPython with pyflyby autoimporter enabled::

    $ py

  Start IPython/Jupyter Notebook with pyflyby autoimporter enabled::

    $ py nb

  Find the ASCII value of the letter "j" (apply builtin function)::

    $ py ord j
    [PYFLYBY] ord('j')
    106

  Decode a base64-encoded string (apply autoimported function)::

    $ py b64decode aGVsbG8=
    [PYFLYBY] from base64 import b64decode
    [PYFLYBY] b64decode('aGVsbG8=', altchars=None)
    b'hello'

  Find the day of the week of some date (apply function in module)::

    $ py calendar.weekday 2014 7 18
    [PYFLYBY] import calendar
    [PYFLYBY] calendar.weekday(2014, 7, 18)
    4

  Using named arguments::

    $ py calendar.weekday --day=16 --month=7 --year=2014
    [PYFLYBY] import calendar
    [PYFLYBY] calendar.weekday(2014, 7, 16)
    2

  Using short named arguments::

    $ py calendar.weekday -m 7 -d 15 -y 2014
    [PYFLYBY] import calendar
    [PYFLYBY] calendar.weekday(2014, 7, 15)
    1

  Invert a matrix (evaluate expression, with autoimporting)::

    $ py 'matrix("1 3 3; 1 4 3; 1 3 4").I'
    [PYFLYBY] from numpy import matrix
    [PYFLYBY] matrix("1 3 3; 1 4 3; 1 3 4").I
    matrix([[ 7., -3., -3.],
            [-1.,  1.,  0.],
            [-1.,  0.,  1.]])

  Plot cosine (evaluate expression, with autoimporting)::

    $ py 'plot(cos(arange(30)))'
    [PYFLYBY] from numpy import arange
    [PYFLYBY] from numpy import cos
    [PYFLYBY] from matplotlib.pyplot import plot
    [PYFLYBY] plot(cos(arange(30)))
    <plot>

  Command-line calculator (multiple arguments)::

    $ py 3 / 4
    0.75

  Command-line calculator (single arguments)::

    $ py '(5+7j) \** 12'
    (65602966976-150532462080j)

  Rationalize a decimal (apply bound method)::

    $ py 2.5.as_integer_ratio
    [PYFLYBY] 2.5.as_integer_ratio()
    (5, 2)

  Rationalize a decimal (apply unbound method)::

    $ py float.as_integer_ratio 2.5
    [PYFLYBY] float.as_integer_ratio(2.5)
    (5, 2)

  Rationalize decimals (map/apply)::

    $ py --map float.as_integer_ratio 2.5 3.5
    [PYFLYBY] float.as_integer_ratio(2.5)
    (5, 2)
    [PYFLYBY] float.as_integer_ratio(3.5)
    (7, 2)

  Square numbers (map lambda)::

    $ py --map 'lambda x: x \**2' 3 4 5
    [PYFLYBY] (lambda x: x \**2)(3)
    9
    [PYFLYBY] (lambda x: x \**2)(4)
    16
    [PYFLYBY] (lambda x: x \**2)(5)
    25

  Find length of string (using "-" for stdin)::

    $ echo hello | py len -
    [PYFLYBY] len('hello\\n')
    6

  Run stdin as code::

    $ echo 'print(sys.argv[1:])' | py - hello world
    [PYFLYBY] import sys
    ['hello', 'world']

  Run libc functions::

    $ py --quiet --output=none 'CDLL("libc.so.6").printf' %03d 7
    007

  Download web page::

    $ py --print 'requests.get(sys.argv[1]).text' http://example.com

  Get function help::

    $ py b64decode?
    [PYFLYBY] from base64 import b64decode
    Python signature::

      >> b64decode(s, altchars=None, validate=False)

    Command-line signature::

      $ py b64decode s [altchars [validate]]
      $ py b64decode --s=... [--altchars=...] [--validate=...]
    ...

  Get module help::

    $ py pandas?
    [PYFLYBY] import pandas
    Version:
      0.13.1
    Filename:
      /usr/local/lib/python2.7/site-packages/pandas/__init__.pyc
    Docstring:
      pandas - a powerful data analysis and manipulation library for Python
      ...

"""



from   functools                import total_ordering
from   pathlib                  import Path
from   typing                   import Any
import warnings

from   pyflyby._util            import cmp
from   shlex                    import quote as shquote


# TODO: add --tidy-imports, etc

# TODO: new --action="concat_eval eval apply" etc.  specifying multiple
# actions means try each of them in that order.  then --safe can exclude
# concat-eval, and users can customize which action modes are included.
# --apply would be equivalent to --action=apply.

# TODO: plug-in system.  'py foo' should attempt something that the
# user/vendor can add to the system.  leading candidate: use entry_point
# system (http://stackoverflow.com/a/774859).  other candidates: maybe use
# python namespace like pyflyby.vendor.foo or pyflyby.commands.foo or
# pyflyby.magics.foo, or maybe a config file.
# TODO: note additional features in documentation feature list

# TODO: somehow do the right thing with glob.glob vs glob, pprint.pprint vs
# pprint, etc.  Ideas:
#   - make --apply special case detect modules and take module.module
#   - enhance auto_import() to keep track of the context while finding missing imports
#   - enhance auto_import() to scan for calls after importing

# TODO: pipe help/source output (all output?) through $PYFLYBY_PAGER (default "less -FRX").

# TODO: unparse ast node for info logging
# https://hg.python.org/cpython/log/tip/Tools/parser/unparse.py

# TODO: run_module should detect if the module doesn't check __name__ and
# therefore is unlikely to be meaningful to use with run_module.

# TODO: make sure run_modules etc work correctly with modules under namespace
# packages.

# TODO: detect deeper ImportError, e.g. suppose user accesses module1; module1
# imports badmodule, which can't be imported successfully and raises
# ImportError; we should get that ImportError instead of trying other things
# or turning it into a string.  probably do this by changing the place where
# we import modules to first get the loader, then if import fails, raise a
# subclass of ImportError.

# TODO: provide a way to omit newline in output.  maybe --output=write.

# TODO: make sure 'py -c' matches behavior of 'python -c' w.r.t. sys.modules["__main__"]
# $ py -c 'x=3;import sys; print sys.modules["__main__"].__dict__["x"]'
# mimic runpy.{_TempModule,_run_code}

# TODO: refactor this module - maybe to _heuristic.py

# TODO: add --profile, --runsnake

import ast
import builtins
from   contextlib               import contextmanager
import inspect
import os
import re
import sys
import types
from   types                    import FunctionType, MethodType

from   pyflyby._autoimp         import auto_import, find_missing_imports
from   pyflyby._cmdline         import print_version_and_exit, syntax
from   pyflyby._dbg             import (add_debug_functions_to_builtins,
                                        attach_debugger, debug_statement,
                                        debugger, enable_faulthandler,
                                        enable_signal_handler_debugger,
                                        enable_sigterm_handler,
                                        remote_print_stack)
from   pyflyby._file            import Filename, UnsafeFilenameError, which
from   pyflyby._flags           import CompilerFlags
from   pyflyby._idents          import is_identifier
from   pyflyby._interactive     import (get_ipython_terminal_app_with_autoimporter,
                                        run_ipython_line_magic,
                                        start_ipython_with_autoimporter)
from   pyflyby._log             import logger
from   pyflyby._modules         import ModuleHandle
from   pyflyby._parse           import PythonBlock
from   pyflyby._util            import indent, prefixes

usage = """
py --- command-line python multitool with automatic importing

$ py [--file]   filename.py arg1 arg2      Execute file
$ py [--apply]  function arg1 arg2         Call function
$ py [--eval]  'function(arg1, arg2)'      Evaluate code
$ py [--module] modname arg1 arg2          Run a module

$ py --debug    file/code... args...       Debug code
$ py --debug    PID                        Attach debugger to PID

$ py                                       IPython shell
""".strip()

# Default compiler flags (feature flags) used for all user code.  We include
# "print_function" here, but we also use auto_flags=True, which means
# print_function may be flipped off if the code contains print statements.
FLAGS = CompilerFlags(["absolute_import", "with_statement", "division",
                       "print_function"])


def _get_argspec(arg:Any) -> inspect.FullArgSpec:
    from inspect import getfullargspec as getargspec, FullArgSpec as ArgSpec
    if isinstance(arg, FunctionType):
        return getargspec(arg)
    elif isinstance(arg, MethodType):
        argspec = getargspec(arg)
        if arg.__self__ is not None:
            # For bound methods, ignore the "self" argument.
            return ArgSpec(argspec.args[1:], *argspec[1:])
        return argspec
    elif isinstance(arg, type):
        if arg.__new__ is not object.__new__:
            argspec = _get_argspec(arg.__new__)
            return ArgSpec(argspec.args[1:], *argspec[1:])
        else:
            argspec = _get_argspec(arg.__init__)
            return ArgSpec(argspec.args[1:], *argspec[1:])
    elif callable(arg):
        # Unknown, probably a built-in method.
        return ArgSpec([], "args", "kwargs", None, [], None, {})
    raise TypeError(
        "_get_argspec: unexpected %s" % (type(arg).__name__,))


def _requires_parens_as_function(function_name:str) -> bool:
    """
    Returns whether the given string of a callable would require parentheses
    around it to call it.

      >>> _requires_parens_as_function("foo.bar[4]")
      False

      >>> _requires_parens_as_function("foo+bar")
      True

      >>> _requires_parens_as_function("(foo+bar)()")
      False

      >>> _requires_parens_as_function("(foo+bar)")
      False

      >>> _requires_parens_as_function("(foo)+(bar)")
      True

    :type function_name:
      ``str``
    :rtype:
      ``bool``
    """
    # TODO: this might be obsolete if we use unparse instead of keeping original
    #     user formatting (or alternatively, unparse should use something like this).

    assert isinstance(function_name, str)

    block = PythonBlock(function_name, flags=FLAGS)
    node = block.expression_ast_node
    if not node:
        # Couldn't parse?  Just assume we do need parens for now.  Or should
        # we raise an exception here?
        return True
    body = node.body
    # Is it something that doesn't need parentheses?
    if isinstance(body, (ast.Name, ast.Attribute, ast.Call, ast.Subscript)):
        return False
    # Does it already have parentheses?
    n = str(function_name)
    if n.startswith("(") and n.endswith(")"):
        # It has parentheses, superficially.  Make sure it's not something
        # like "(foo)+(bar)".
        flags = int(FLAGS) | ast.PyCF_ONLY_AST
        try:
            tnode = compile(n[1:-1], "<unknown>", "eval", flags)
        except SyntaxError:
            return True
        if ast.dump(tnode) == ast.dump(node):
            return False
        else:
            return True
    return True


def _format_call_spec(function_name:str, obj:Any)-> str:
    # using signature() is not strictly identical
    # as it might look at __text_signature__ and/or respect possitional only
    # forward slash:
    #   >>> def foo(a, /, b)
    # whcih formatargspec did not do.
    # argspec = _get_argspec(obj)
    # old_callspect = inspect.formatargspec(*argspec)
    # assert old_callspect == callspec , f"{old_callspect} !={callspec}"
    callspec = str(inspect.signature(obj))
    if _requires_parens_as_function(function_name):
        return "(%s)%s" % (function_name, callspec)
    else:
        return "%s%s" % (function_name, callspec)




def _build_function_usage_string(function_name:str, obj:Any, prefix:str):
    argspec = _get_argspec(obj)

    usage = []
    # TODO: colorize
    usage.append("Python signature:")
    usage.append("  >"+">> " + _format_call_spec(function_name, obj))
    usage.append("")
    usage.append("Command-line signature:")
    keywords = argspec.varkw
    if not argspec.args and argspec.varargs and keywords:
        # We have no information about the arguments.  It's probably a
        # built-in where getargspec failed.
        usage.append("  $ %s%s ...\n" % (prefix, function_name))
        return "\n".join(usage)
    defaults = argspec.defaults or ()
    first_with_default = len(argspec.args) - len(defaults)
    # Show first alternative of command-line syntax.
    syntax1 = "  $ %s%s" % (prefix, shquote(function_name),)
    for i, arg in enumerate(argspec.args):
        if i >= first_with_default:
            syntax1 += " [%s" % (arg,)
        else:
            syntax1 += " %s" % (arg,)
    if argspec.varargs:
        syntax1 += " %s..." % argspec.varargs
    syntax1 += "]" * len(defaults)
    for arg in argspec.kwonlyargs:
        if argspec.kwonlydefaults and arg in argspec.kwonlydefaults:
            syntax1 += " [--%s=...]" % (arg,)
        else:
            syntax1 += " --%s=..." % (arg,)
    if keywords:
        syntax1 += " [--...]"
    usage.append(syntax1)
    # usage.append("or:")
    syntax2 = "  $ %s%s" % (prefix, shquote(function_name),)
    for i, arg in enumerate(argspec.args):
        if i >= first_with_default:
            syntax2 += " [--%s=...]" % (arg,)
        else:
            syntax2 += " --%s=..." % (arg,)
    for arg in argspec.kwonlyargs:
        if argspec.kwonlydefaults and arg in argspec.kwonlydefaults:
            syntax2 += " [--%s=...]" % (arg,)
        else:
            syntax2 += " --%s=..." % (arg,)
    if argspec.varargs:
        syntax2 += " %s..." % argspec.varargs
    if keywords:
        syntax2 += " [--...]"
    usage.append(syntax2)
    usage.append("")
    return "\n".join(usage)


class ParseError(Exception):
    pass


class _ParseInterruptedWantHelp(Exception):
    pass


class _ParseInterruptedWantSource(Exception):
    pass


class UserExpr:
    """
    An expression from user input, and its evaluated value.

    The expression can come from a string literal or other raw value, or a
    string that is evaluated as an expression, or heuristically chosen.

      >>> ns = _Namespace()

    Heuristic auto-evaluation::

      >>> UserExpr('5+2', ns, "auto").value
      7

      >>> UserExpr('5j+2', ns, "auto").value
      (2+5j)

      >>> UserExpr('base64.b64decode("SGFsbG93ZWVu")', ns, "auto").value
      [PYFLYBY] import base64
      b'Halloween'

    Returning an unparsable argument as a string::

      >>> UserExpr('Victory Loop', ns, "auto").value
      'Victory Loop'

    Returning an undefined (and not auto-importable) argument as a string::

      >>> UserExpr('Willowbrook29817621+5', ns, "auto").value
      'Willowbrook29817621+5'

    Explicit literal string::

      >>> UserExpr("2+3", ns, "raw_value").value
      '2+3'

      >>> UserExpr("'2+3'", ns, "raw_value").value
      "'2+3'"

    Other raw values::

      >>> UserExpr(sys.exit, ns, "raw_value").value
      <built-in function exit>
    """

    def __init__(self, arg, namespace, arg_mode, source=None):
        """
        Construct a new UserExpr.

        :type arg:
          ``str`` if ``arg_mode`` is "eval" or "auto"; anything if ``arg_mode``
          is "raw_value"
        :param arg:
          Input user argument.
        :type namespace:
          `_Namespace`
        :type arg_mode:
          ``str``
        :param arg_mode:
          If ``"raw_value"``, then return ``arg`` unchanged.  If ``"eval"``, then
          always evaluate ``arg``.  If ``"auto"``, then heuristically evaluate
          if appropriate.
        """
        if arg_mode == "string":
            arg_mode = "raw_value"
        self._namespace = namespace
        self._original_arg_mode = arg_mode
        self._original_arg        = arg
        if arg_mode == "raw_value":
            # self.inferred_arg_mode = "raw_value"
            # self.original_source = None
            if source is None:
                source = PythonBlock(repr(self._original_arg))
            else:
                source = PythonBlock(source)
            self.source = source
            self.value = self._original_arg
        elif arg_mode == "eval":
            if source is not None:
                raise ValueError(
                    "UserExpr(): source argument not allowed for eval")
            # self.inferred_arg_mode = "eval"
            self._original_arg_as_source = PythonBlock(arg, flags=FLAGS)
            # self.original_source = self._original_arg_as_source
        elif arg_mode == "auto":
            if source is not None:
                raise ValueError(
                    "UserExpr(): source argument not allowed for auto")
            if not isinstance(arg, str):
                raise ValueError(
                    "UserExpr(): arg must be a string if arg_mode='auto'")
            self._original_arg_as_source = PythonBlock(arg, flags=FLAGS)
        else:
            raise ValueError("UserExpr(): bad arg_mode=%r" % (arg_mode,))

    def _infer_and_evaluate(self) -> None:
        if self._original_arg_mode == "raw_value":
            pass
        elif self._original_arg_mode == "eval":
            block = self._original_arg_as_source
            if not (str(block).strip()):
                raise ValueError("empty input")
            self.value = self._namespace.auto_eval(block)
            self.source = self._original_arg_as_source #.pretty_print() # TODO
        elif self._original_arg_mode == "auto":
            block = self._original_arg_as_source
            ERROR = object()
            if not (str(block).strip()):
                value = ERROR
            elif not block.parsable_as_expression:
                value = ERROR
            else:
                try:
                    value = self._namespace.auto_eval(block)
                except UnimportableNameError:
                    value = ERROR
            if value is ERROR:
                # self.inferred_arg_mode = "raw_value"
                self.value = self._original_arg
                # self.original_source = None
                self.source = PythonBlock(repr(self.value))
            else:
                # self.inferred_arg_mode = "eval"
                self.value = value
                # self.original_source = block
                self.source = block #.pretty_print() # TODO
        else:
            raise AssertionError("internal error")
        self._infer_and_evaluate = lambda: None

    def __getattr__(self, k):
        self._infer_and_evaluate()
        return object.__getattribute__(self, k)

    def __str__(self):
        return str(self._original_arg)


def _parse_auto_apply_args(argspec, commandline_args, namespace, arg_mode="auto"):
    """
    Parse command-line arguments heuristically.  Arguments that can be
    evaluated are evaluated; otherwise they are treated as strings.

    :returns:
      ``args``, ``kwargs``
    """
    # This is implemented manually instead of using optparse or argparse.  We
    # do so because neither supports dynamic keyword arguments well.  Optparse
    # doesn't support parsing known arguments only, and argparse doesn't
    # support turning off interspersed positional arguments.
    def make_expr(arg, arg_mode=arg_mode):
        return UserExpr(arg, namespace, arg_mode)

    # Create a map from argname to default value.
    defaults = argspec.defaults or ()
    argname2default = {}
    for argname, default in zip(argspec.args[len(argspec.args)-len(defaults):],
                                defaults):
        argname2default[argname] = make_expr(default, "raw_value")
    if argspec.kwonlydefaults:
        for argname, default in argspec.kwonlydefaults.items():
            argname2default[argname] = make_expr(default, "raw_value")
    # Create a map from prefix to arguments with that prefix.  E.g. {"foo":
    # ["foobar", "foobaz"]}
    prefix2argname = {}
    for argname in argspec.args:
        for prefix in prefixes(argname):
            prefix2argname.setdefault(prefix, []).append(argname)
    for argname in argspec.kwonlyargs:
        for prefix in prefixes(argname):
            prefix2argname.setdefault(prefix, []).append(argname)
    # Enumerate over input arguments.
    got_pos_args = []
    got_keyword_args = {}
    args = list(commandline_args)
    while args:
        arg = args.pop(0)
        if arg in ["--?", "-?", "?"]:
            raise _ParseInterruptedWantHelp
        elif arg in ["--??", "-??", "??"]:
            raise _ParseInterruptedWantSource
        elif arg.startswith("-"):
            if arg == "-":
                # Read from stdin and stuff into next argument as a string.
                data = sys.stdin.read()
                got_pos_args.append(make_expr(data, "string"))
                continue
            elif arg == "--":
                # Treat remaining arguments as strings.
                got_pos_args.extend([make_expr(x, "string") for x in args])
                del args[:]
                continue
            elif arg.startswith("--"):
                argname = arg[2:]
            else:
                argname = arg[1:]
            argname, equalsign, value = argname.partition("=")
            argname = argname.replace("-", "_")
            if not is_identifier(argname):
                raise ParseError("Invalid option name %s" % (argname,))
            matched_argnames = prefix2argname.get(argname, [])
            if len(matched_argnames) == 1:
                argname, = matched_argnames
            elif len(matched_argnames) == 0:
                if equalsign == "":
                    if argname in ["help", "h"]:
                        raise _ParseInterruptedWantHelp
                    if argname in ["source"]:
                        raise _ParseInterruptedWantSource
                if not argspec.varkw:
                    raise ParseError("Unknown option name %s" %
                                     (argname,))

            elif len(matched_argnames) > 1:
                raise ParseError(
                    "Ambiguous %s: could mean one of: %s"
                    % (argname,
                       ", ".join("--%s"%s for s in matched_argnames)))
            else:
                raise AssertionError
            if not value:
                try:
                    value = args.pop(0)
                except IndexError:
                    raise ParseError("Missing argument to %s" % (arg,))
                if value.startswith("--"):
                    raise ParseError(
                        "Missing argument to %s.  "
                        "If you really want to use %r as the argument to %s, "
                        "then use %s=%s."
                        % (arg, value, arg, arg, value))
            got_keyword_args[argname] = make_expr(value)
        else:
            got_pos_args.append(make_expr(arg))

    parsed_args = []
    parsed_kwargs = {}
    for i, argname in enumerate(argspec.args):
        if i < len(got_pos_args):
            if argname in got_keyword_args:
                raise ParseError(
                    "%s specified both as positional argument (%s) "
                    "and keyword argument (%s)"
                    % (argname, got_pos_args[i], got_keyword_args[argname]))
            expr = got_pos_args[i]
        else:
            try:
                expr = got_keyword_args.pop(argname)
            except KeyError:
                try:
                    expr = argname2default[argname]
                except KeyError:
                    raise ParseError(
                        "missing required argument %s" % (argname,))
        try:
            value = expr.value
        except Exception as e:
            raise ParseError(
                "Error parsing value for --%s=%s: %s: %s"
                % (argname, expr, type(e).__name__, e))
        parsed_args.append(value)

    for argname in argspec.kwonlyargs:
        try:
            expr = got_keyword_args.pop(argname)
        except KeyError:
            try:
                expr = argname2default[argname]
            except KeyError:
                raise ParseError(
                    "missing required keyword argument %s" % (argname,))

        try:
            value = expr.value
        except Exception as e:
            raise ParseError(
                "Error parsing value for --%s=%s: %s: %s"
                % (argname, expr, type(e).__name__, e))
        parsed_kwargs[argname] = value

    if len(got_pos_args) > len(argspec.args):
        if argspec.varargs:
            for expr in got_pos_args[len(argspec.args):]:
                try:
                    value = expr.value
                except Exception as e:
                    raise ParseError(
                        "Error parsing value for *%s: %s: %s: %s"
                        % (argspec.varargs, expr, type(e).__name__, e))
                parsed_args.append(value)
        else:
            max_nargs = len(argspec.args)
            if argspec.defaults:
                expected = "%d-%d" % (max_nargs-len(argspec.defaults),max_nargs)
            else:
                expected = "%d" % (max_nargs,)
            raise ParseError(
                "Too many positional arguments.  "
                "Expected %s positional argument(s): %s.  Got %d args: %s"
                % (expected, ", ".join(argspec.args),
                   len(got_pos_args), " ".join(map(str, got_pos_args))))

    for argname, expr in sorted(got_keyword_args.items()):
        try:
            parsed_kwargs[argname] = expr.value
        except Exception as e:
            raise ParseError(
                "Error parsing value for --%s=%s: %s: %s"
                % (argname, expr, type(e).__name__, e))

    return parsed_args, parsed_kwargs


def _format_call(function_name:str, argspec, args, kwargs):
    # TODO: print original unparsed arg strings
    defaults = argspec.defaults or ()
    first_with_default = len(argspec.args) - len(defaults)
    argparts = []
    for i in range(max(len(args), len(argspec.args))):
        if i >= first_with_default and len(args) <= len(argspec.args):
            argparts.append("%s=%r" % (argspec.args[i], args[i]))
        else:
            argparts.append(repr(args[i]))
    for k, v in sorted(kwargs.items()):
        argparts.append("%s=%r" % (k, v))
    if _requires_parens_as_function(function_name):
        function_name = "(%s)" % (function_name,)
    r = "%s(%s)" % (function_name, ", ".join(argparts))
    return r


class UnimportableNameError(NameError):
    pass


class NotAFunctionError(Exception):
    pass


def _get_help(expr:UserExpr, verbosity:int=1) -> str:
    """
    Construct a help string.

    :type expr:
      `UserExpr`
    :param expr:
      Object to generate help for.
    :rtype:
      ``str``
    """
    # TODO: colorize headers
    result = ""
    obj = expr.value
    name:str = str(expr.source)
    if callable(obj):
        prefix = os.path.basename(sys.orig_argv[0]) + " "
        result += _build_function_usage_string(name, obj, prefix)
    if verbosity == 0:
        include_filename = False
        include_doc      = False
        include_source   = False
    elif verbosity == 1:
        include_filename = True
        include_doc      = True
        include_source   = False
    elif verbosity == 2:
        include_filename = True
        include_doc      = False
        include_source   = True
    else:
        raise ValueError("invalid verbosity=%r" % (verbosity,))
    try:
        version = obj.__version__
    except Exception:
        pass
    else:
        result += "\nVersion:\n  %s\n" % (version,)
    if include_filename:
        try:
            filename = inspect.getfile(obj)
        except Exception:
            pass
        else:
            result += "\nFilename:\n  %s\n" % (filename,)
    if include_source:
        try:
            source = inspect.getsource(obj)
        except Exception:
            source = ""
        if source:
            # TODO: colorize source
            result += "\nSource:\n%s\n" % (indent(source, "  "))
        else:
            source = "(Not available)"
            include_doc = True
    if include_doc:
        doc = (inspect.getdoc(obj) or "").strip() or "(No docstring)"
        result += "\nDocstring:\n%s" % (indent(doc, "  "))
    return result


_enable_postmortem_debugger = None

def _handle_user_exception(exc_info=None):
    """
    Handle an exception in user code.
    """
    # TODO: Make tracebacks show user code being executed.  IPython does that
    # by stuffing linecache.cache, and also advising linecache.checkcache.  We
    # can either advise it ourselves also, or re-use IPython.core.compilerop.
    # Probably better to advise ourselves.  Add "<pyflyby-input-md5[:12]>" to
    # linecache.  Perhaps do it in a context manager that removes from
    # linecache when done.  Advise checkcache to protect any "<pyflyby-*>".
    # Do it for all code compiled from here, including args, debug_statement,
    # etc.
    if exc_info is None:
        exc_info = sys.exc_info()
    if exc_info[2].tb_next:
        exc_info = (exc_info[0], exc_info[1],
                    exc_info[2].tb_next) # skip this traceback
    # If ``_enable_postmortem_debugger`` is enabled, then debug the exception.
    # By default, this is enabled run running in a tty.
    # We check isatty(1) here because we want 'py ... | cat' to never go into
    # the debugger.  Note that debugger() also checks whether /dev/tty is
    # usable (and if not, waits for attach).
    if _enable_postmortem_debugger:
        # *** Run debugger. ***
        debugger(exc_info)
    # TODO: consider using print_verbose_tb(*exc_info)
    import traceback
    traceback.print_exception(*exc_info)
    raise SystemExit(1)


def auto_apply(function, commandline_args, namespace, arg_mode=None,
               debug=False):
    """
    Call ``function`` on command-line arguments.  Arguments can be positional
    or keyword arguments like "--foo=bar".  Arguments are by default
    heuristically evaluated.

    :type function:
      ``UserExpr``
    :param function:
      Function to apply.
    :type commandline_args:
      ``list`` of ``str``
    :param commandline_args:
      Arguments to ``function`` as strings.
    :param arg_mode:
      How to interpret ``commandline_args``.  If ``"string"``, then treat them
      as literal strings.  If ``"eval"``, then evaluate all arguments as
      expressions.  If ``"auto"`` (the default), then heuristically decide
      whether to treat as expressions or strings.
    """
    if not isinstance(function, UserExpr):
        raise TypeError
    if not callable(function.value):
        raise NotAFunctionError("Not a function", function.value)
    arg_mode = _interpret_arg_mode(arg_mode, default="auto")
    # Parse command-line arguments.
    argspec = _get_argspec(function.value)
    try:
        args, kwargs = _parse_auto_apply_args(argspec, commandline_args,
                                              namespace, arg_mode=arg_mode)
    except _ParseInterruptedWantHelp:
        usage = _get_help(function, verbosity=1)
        print(usage)
        raise SystemExit(0)
    except _ParseInterruptedWantSource:
        usage = _get_help(function, verbosity=2)
        print(usage)
        raise SystemExit(0)
    except ParseError as e:
        # Failed parsing command-line arguments.  Print usage.
        logger.error(e)
        usage = _get_help(function, verbosity=(1 if logger.info_enabled else 0))
        sys.stderr.write("\n" + usage)
        raise SystemExit(1)
    # Log what we're doing.

    logger.info("%s", _format_call(str(function.source), argspec, args, kwargs))

    # *** Call the function. ***
    f = function.value
    try:
        if debug:
            result = debug_statement("f(*args, **kwargs)")
        else:
            result = f(*args, **kwargs)
        return result
    except SystemExit:
        raise
    # TODO: handle "quit" by user here specially instead of returning None.
    # May need to reimplement pdb.runeval() so we can catch BdbQuit.
    except:
        # Handle exception in user code.
        _handle_user_exception()


@total_ordering
class LoggedList:
    """
    List that logs which members have not yet been accessed (nor removed).
    """

    _ACCESSED = object()

    def __init__(self, items):
        self._items = list(items)
        self._unaccessed = list(self._items)

    def append(self, x):
        self._unaccessed.append(self._ACCESSED)
        self._items.append(x)

    def count(self):
        return self._items.count()

    def extend(self, new_items):
        new_items = list(new_items)
        self._unaccessed.extend([self._ACCESSED] * len(new_items))
        self._items.extend(new_items)

    def index(self, x, *start_stop):
        index = self._items.index(x, *start_stop) # may raise ValueError
        self._unaccessed[index] = self._ACCESSED
        return index

    def insert(self, index, x):
        self._unaccessed.insert(index, self._ACCESSED)
        self._items.insert(index, x)

    def pop(self, index):
        self._unaccessed.pop(index)
        return self._items.pop(index)

    def remove(self, x):
        index = self._items.index(x)
        self.pop(index)

    def reverse(self):
        self._items.reverse()
        self._unaccessed.reverse()

    def sort(self):
        indexes = range(len(self._items))
        indexes.sort(key=self._items.__getitem__) # argsort
        self._items = [self._items[i] for i in indexes]
        self._unaccessed = [self._unaccessed[i] for i in indexes]

    def __add__(self, other):
        return self._items + other

    def __contains__(self, x):
        try:
            self.index(x)
            return True
        except ValueError:
            return False

    def __delitem__(self, x):
        del self._items[x]
        del self._unaccessed[x]

    def __eq__(self, other):
        if not isinstance(other, LoggedList):
            return NotImplemented
        return self._items == other._items

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if not isinstance(other, LoggedList):
            return NotImplemented
        return self._items < other._items

    def __cmp__(self, x):
        return cmp(self._items, x)

    def __getitem__(self, idx):
        result = self._items[idx]
        if isinstance(idx, slice):
            self._unaccessed[idx] = [self._ACCESSED]*len(result)
        else:
            self._unaccessed[idx] = self._ACCESSED
        return result

    def __hash__(self):
        raise TypeError("unhashable type: 'LoggedList'")

    def __iadd__(self, x):
        self.extend(x)

    def __imul__(self, n):
        self._items *= n
        self._unaccessed *= n

    def __iter__(self):
        # Todo: detect mutation while iterating.
        for i, x in enumerate(self._items):
            self._unaccessed[i] = self._ACCESSED
            yield x

    def __len__(self):
        return len(self._items)


    def __mul__(self, n):
        return self._items * n

    def __reduce__(self):
        return

    def __repr__(self):
        self._unaccessed[:] = [self._ACCESSED]*len(self._unaccessed)
        return repr(self._items)

    def __reversed__(self):
        # Todo: detect mutation while iterating.
        for i in reversed(range(len(self._items))):
            self._unaccessed[i] = self._ACCESSED
            yield self._items[i]

    def __rmul__(self, n):
        return self._items * n

    def __setitem__(self, idx, value):
        self._items[idx] = value
        if isinstance(idx, slice):
            self._unaccessed[idx] = [self._ACCESSED]*len(value)
        else:
            self._unaccessed[idx] = value

    def __str__(self):
        self._unaccessed[:] = [self._ACCESSED]*len(self._unaccessed)
        return str(self._items)

    @property
    def unaccessed(self):
        return [x for x in self._unaccessed if x is not self._ACCESSED]


@contextmanager
def SysArgvCtx(*args):
    """
    Context manager that:
      * Temporarily sets sys.argv = args.
      * At end of context, complains if any args were never accessed.
    """
    # There should always be at least one arg, since the first one is
    # the program name.
    if not args:
        raise ValueError("Missing args")
    nargs = len(args) - 1
    # Create a list proxy that will log accesses.
    argv = LoggedList(args)
    # Don't consider first argument to be interesting.
    argv[0]
    orig_argv = list(sys.argv)
    try:
        sys.argv = argv
        # Run context code.
        yield
        # Complain if there are unaccessed arguments.
        unaccessed = argv.unaccessed
        if not unaccessed:
            pass
        else:
            if nargs == 1:
                msg = ("You specified a command-line argument, but your code didn't use it: %s"
                       % (unaccessed[0],))
            elif len(unaccessed) == nargs:
                msg = ("You specified %d command-line arguments, but your code didn't use them: %s"
                       % (len(unaccessed), " ".join(unaccessed)))
            else:
                msg = ("You specified %d command-line arguments, but your code didn't use %d of them: %s"
                       % (nargs, len(unaccessed), " ".join(unaccessed)))
            msg2 = "\nIf this is intentional, access 'sys.argv[:]' somewhere in your code."
            logger.error(msg + msg2)
            raise SystemExit(1)
    finally:
        sys.argv = orig_argv


def _as_filename_if_seems_like_filename(arg):
    """
    If ``arg`` seems like a filename, then return it as one.

      >>> bool(_as_filename_if_seems_like_filename("foo.py"))
      True

      >>> bool(_as_filename_if_seems_like_filename("%foo.py"))
      False

      >>> bool(_as_filename_if_seems_like_filename("foo(bar)"))
      False

      >>> bool(_as_filename_if_seems_like_filename("/foo/bar/baz.quux-660470"))
      True

      >>> bool(_as_filename_if_seems_like_filename("../foo/bar-24084866"))
      True

    :type arg:
      ``str``
    :rtype:
      ``Filename``
    """
    try:
        filename = Filename(arg)
    except UnsafeFilenameError:
        # If the filename isn't a "safe" filename, then don't treat it as one,
        # and don't even check whether it exists.  This means that for an
        # argument like "foo(bar)" or "lambda x:x*y" we won't even check
        # existence.  This is both a performance optimization and a safety
        # valve to avoid unsafe filenames being created to intercept expressions.
        return None
    # If the argument "looks" like a filename and is unlikely to be a python
    # expression, then assume it is a filename.  We assume so regardless of
    # whether the file actually exists; if it turns out to not exist, we'll
    # complain later.
    if arg.startswith("/") or arg.startswith("./") or arg.startswith("../"):
        return filename
    if filename.ext == ".py":
        # TODO: .pyc, .pyo
        return which(arg) or filename
    # Even if it doesn't obviously look like a filename, but it does exist as
    # a filename, then use it as one.
    if filename.exists:
        return filename
    # If it's a plain name and we can find an executable on $PATH, then use
    # that.
    if re.match("^[a-zA-Z0-9_-]+$", arg):
        filename = which(arg)
        if not filename:
            return None
        if not _has_python_shebang(filename):
            logger.debug("Found %s but it doesn't seem like a python script",
                         filename)
            return None
        return filename
    return None


def _has_python_shebang(filename):
    """
    Return whether the first line contains #!...python...

    Used for confirming that an executable script found via which() is
    actually supposed to be a python script.

    Note that this test is only needed for scripts found via which(), since
    otherwise the shebang is not necessary.
    """
    assert isinstance(filename, Filename)
    with open(str(filename), 'rb') as f:
        line = f.readline(1024)
        return line.startswith(b"#!") and b"python" in line



def _interpret_arg_mode(arg, default="auto"):
    """
      >>> _interpret_arg_mode("Str")
      'string'
    """
    if arg is None:
        arg = default
    if arg == "auto" or arg == "eval" or arg == "string":
        return arg # optimization for interned strings
    rarg = str(arg).strip().lower()
    if rarg in ["eval", "evaluate", "exprs", "expr", "expressions", "expression", "e"]:
        return "eval"
    elif rarg in ["strings", "string", "str", "strs", "literal", "literals", "s"]:
        return "string"
    elif rarg in ["auto", "automatic", "a"]:
        return "auto"
    elif rarg == "error":
        # Intentionally not documented to user
        return "error"
    else:
        raise ValueError(
            "Invalid arg_mode=%r; expected one of eval/string/auto"
            % (arg,))


def _interpret_output_mode(arg, default="interactive"):
    """
      >>> _interpret_output_mode('Repr_If_Not_None')
      'repr-if-not-none'
    """
    if arg is None:
        arg = default
    rarg = str(arg).strip().lower().replace("-", "").replace("_", "")
    if rarg in ["none", "no", "n", "silent"]:
        return "silent"
    elif rarg in ["interactive", "i"]:
        return "interactive"
    elif rarg in ["print", "p", "string", "str"]:
        return "str"
    elif rarg in ["repr", "r"]:
        return "repr"
    elif rarg in ["pprint", "pp"]:
        return "pprint"
    elif rarg in ["reprifnotnone", "reprunlessnone", "rn"]:
        return "repr-if-not-none"
    elif rarg in ["pprintifnotnone", "pprintunlessnone", "ppn"]:
        return "pprint-if-not-none"
    elif rarg in ["systemexit", "exit", "raise"]:
        return "exit"
    else:
        raise ValueError(
            "Invalid output=%r; expected one of "
            "silent/interactive/str/repr/pprint/repr-if-not-none/pprint-if-not-none/exit"
            % (arg,))


def print_result(result, output_mode):
    output_mode = _interpret_output_mode(output_mode)
    if output_mode == "silent":
        return
    if output_mode == "interactive":
        # TODO: support IPython output stuff (text/plain)
        try:
            idisp = result.__interactive_display__
        except Exception:
            output_mode = "pprint-if-not-none"
        else:
            result = idisp()
            output_mode = "print-if-not-none"
        # Fall through.
    if output_mode == "str":
        print(str(result))
    elif output_mode == "repr":
        print(repr(result))
    elif output_mode == "pprint":
        import pprint
        pprint.pprint(result) # or equivalently, print pprint.pformat(result)
    elif output_mode == "repr-if-not-none":
        if result is not None:
            print(repr(result))
    elif output_mode == "print-if-not-none":
        if result is not None:
            print(result)
    elif output_mode == "pprint-if-not-none":
        if result is not None:
            import pprint
            pprint.pprint(result)
    elif output_mode == "exit":
        # TODO: only raise at end after pre_exit
        raise SystemExit(result)
    else:
        raise AssertionError("unexpected output_mode=%r" % (output_mode,))


def _get_path_links(p: Path):
    """Gets path links including all symlinks.

    Adapted from `IPython.core.interactiveshell.InteractiveShell.get_path_links`.
    """
    paths = [p]
    while p.is_symlink():
        new_path = Path(os.readlink(p))
        if not new_path.is_absolute():
            new_path = p.parent / new_path
        p = new_path
        paths.append(p)
    return paths


def _init_virtualenv():
    """Add the current virtualenv to sys.path so the user can import modules from it.

    A warning will appear suggesting the user installs IPython in the
    virtualenv, but for many cases, it probably works well enough.

    Adapted `IPython.core.interactiveshell.InteractiveShell.init_virtualenv`.
    """
    if 'VIRTUAL_ENV' not in os.environ:
        # Not in a virtualenv
        return
    elif os.environ["VIRTUAL_ENV"] == "":
        warnings.warn("Virtual env path set to '', please check if this is intended.")
        return

    p = Path(sys.executable)
    p_venv = Path(os.environ["VIRTUAL_ENV"]).resolve()

    # fallback venv detection:
    # stdlib venv may symlink sys.executable, so we can't use realpath.
    # but others can symlink *to* the venv Python, so we can't just use sys.executable.
    # So we just check every item in the symlink tree (generally <= 3)
    paths = _get_path_links(p)

    # In Cygwin paths like "c:\..." and '\cygdrive\c\...' are possible
    if len(p_venv.parts) > 2 and p_venv.parts[1] == "cygdrive":
        drive_name = p_venv.parts[2]
        p_venv = (drive_name + ":/") / Path(*p_venv.parts[3:])

    if any(p_venv == p.parents[1].resolve() for p in paths):
        # Our exe is inside or has access to the virtualenv, don't need to do anything.
        return

    if sys.platform == "win32":
        virtual_env = str(Path(os.environ["VIRTUAL_ENV"], "Lib", "site-packages"))
    else:
        virtual_env_path = Path(
            os.environ["VIRTUAL_ENV"], "lib", "python{}.{}", "site-packages"
        )
        p_ver = sys.version_info[:2]

        # Predict version from py[thon]-x.x in the $VIRTUAL_ENV
        re_m = re.search(r"\bpy(?:thon)?([23])\.(\d+)\b", os.environ["VIRTUAL_ENV"])
        if re_m:
            predicted_path = Path(str(virtual_env_path).format(*re_m.groups()))
            if predicted_path.exists():
                p_ver = re_m.groups()

        virtual_env = str(virtual_env_path).format(*p_ver)

    warnings.warn(
        "Attempting to work in a virtualenv. If you encounter problems, "
        "please install pyflyby inside the virtualenv."
    )
    import site
    sys.path.insert(0, virtual_env)
    site.addsitedir(virtual_env)


class _Namespace(object):

    def __init__(self):
        _init_virtualenv()
        self.globals      = {"__name__": "__main__",
                             "__builtin__": builtins,
                             "__builtins__": builtins}
        self.autoimported = {}

    def auto_import(self, arg):
        return auto_import(arg, [self.globals], autoimported=self.autoimported)

    def auto_eval(self, block, mode=None, info=False, auto_import=True,
                  debug=False):
        """
        Evaluate ``block`` with auto-importing.
        """
        # Equivalent to::
        #   auto_eval(arg, mode=mode, flags=FLAGS, globals=self.globals)
        # but better logging and error raising.
        if not isinstance(block, PythonBlock):
            block = PythonBlock(block, flags=FLAGS, auto_flags=True)
        if auto_import and not self.auto_import(block):
            missing = find_missing_imports(block, [self.globals])
            mstr = ", ".join(repr(str(x)) for x in missing)
            if len(missing) == 1:
                msg = "name %s is not defined and not importable" % mstr
            elif len(missing) > 1:
                msg = "names %s are not defined and not importable" % mstr
            else:
                raise AssertionError
            raise UnimportableNameError(msg)
        if info:
            logger.info(block)
        try:
            # TODO: enter text into linecache
            if debug:
                return debug_statement(block, self.globals)
            else:
                code = block.compile(mode=mode)
                return eval(code, self.globals, self.globals)
        except SystemExit:
            raise
        except:
            _handle_user_exception()

    def __repr__(self):
        return "<{} object at 0x{:0x} \nglobals:{} \nautoimported:{}>".format(
            type(self).__name__, id(self), self.globals, self.autoimported
        )


class _PyMain(object):

    def __init__(self, args):
        self.main_args = args
        self.namespace = _Namespace()
        self.result = None
        self.ipython_app = None

    def exec_stdin(self, cmd_args):
        arg_mode = _interpret_arg_mode(self.arg_mode, default="string")
        output_mode = _interpret_output_mode(self.output_mode, default="silent")
        cmd_args = [UserExpr(a, self.namespace, arg_mode).value
                    for a in cmd_args]
        with SysArgvCtx(*cmd_args):
            result = self.namespace.auto_eval(Filename.STDIN, debug=self.debug)
            print_result(result, output_mode)
            self.result = result

    def eval(self, cmd, cmd_args):
        arg_mode = _interpret_arg_mode(self.arg_mode, default="string")
        output_mode = _interpret_output_mode(self.output_mode)
        cmd_args = [UserExpr(a, self.namespace, arg_mode).value
                    for a in cmd_args]
        with SysArgvCtx("-c", *cmd_args):
            cmd = PythonBlock(cmd)
            result = self.namespace.auto_eval(cmd, info=True, debug=self.debug)
            # TODO: make auto_eval() plow ahead even if there are unimportable
            # names, after warning
            print_result(result, output_mode)
            self.result = result

    def execfile(self, filename_arg, cmd_args):
        # TODO: pass filename to import db target_filename; unit test.
        # TODO: set __file__
        # TODO: support compiled (.pyc/.pyo) files
        arg_mode = _interpret_arg_mode(self.arg_mode, default="string")
        output_mode = _interpret_output_mode(self.output_mode)
        cmd_args = [UserExpr(a, self.namespace, arg_mode).value
                    for a in cmd_args]
        additional_msg = ""
        if isinstance(filename_arg, Filename):
            filename = filename_arg
        elif filename_arg == "-":
            filename = Filename.STDIN
        elif "/" in filename_arg:
            filename = Filename(filename_arg)
        else:
            filename = which(filename_arg)
            if not filename:
                filename = Filename(filename_arg)
                additional_msg = (" (and didn't find %s on $PATH)"
                                  % (filename_arg,))
            elif not _has_python_shebang(filename):
                additional_msg = (" (found %s but it doesn't look "
                                  "like python source"
                                  % (filename,))
                filename = Filename(filename_arg)
        if not filename.exists:
            raise Exception("No such file: %s%s" % (filename, additional_msg))
        with SysArgvCtx(str(filename), *cmd_args):
            sys.path.insert(0, str(filename.dir))
            self.namespace.globals["__file__"] = str(filename)
            result = self.namespace.auto_eval(filename, debug=self.debug)
            print_result(result, output_mode)
            self.result = result

    def apply(self, function, cmd_args):
        arg_mode = _interpret_arg_mode(self.arg_mode, default="auto")
        output_mode = _interpret_output_mode(self.output_mode)
        # Todo: what should we set argv to?
        result = auto_apply(function, cmd_args, self.namespace, arg_mode,
                            debug=self.debug)
        print_result(result, output_mode)
        self.result = result

    def _seems_like_runnable_module(self, arg):
        if not is_identifier(arg, dotted=True):
            # It's not a single (dotted) identifier.
            return False
        if not find_missing_imports(arg, [{}]):
            # It's off of a builtin, e.g. "str.upper"
            return False
        m = ModuleHandle(arg)
        if m.parent:
            # Auto-import the parent, which is necessary in order to get the
            # filename of the module.  ``ModuleHandle.filename`` does this
            # automatically, but we do it explicitly here so that we log
            # the import of the parent module.
            if not self.namespace.auto_import(str(m.parent.name)):
                return False
        if not m.filename:
            logger.debug("Module %s doesn't have a source filename", m)
            return False
        # TODO: check that the source accesses __main__ (ast traversal?)
        return True

    def heuristic_cmd(self, cmd, cmd_args, function_name=None):
        output_mode = _interpret_output_mode(self.output_mode)
        # If the "command" is just a module name, then call run_module.  Make
        # sure we check that it's not a builtin.
        if self._seems_like_runnable_module(str(cmd)):
            self.heuristic_run_module(str(cmd), cmd_args)
            return
        # FIXME TODO heed arg_mode for non-apply case.  This is tricky to
        # implement; will require assigning some proxy class to sys.argv
        # that's more sophisticated than just logging.
        with SysArgvCtx("-c", *cmd_args):
            # Log the expression before we evaluate it, unless we're likely to
            # log another substantially similar line.  (We can only guess
            # heuristically whether it's interesting enough to log it.  And we
            # can't know whether the result will be callable until we evaluate
            # it.)
            info = not re.match("^[a-zA-Z0-9_.]+$", function_name)
            result = self.namespace.auto_eval(cmd, info=info, debug=self.debug)
            if callable(result):
                function = UserExpr(
                    result, self.namespace, "raw_value", function_name)
                result = auto_apply(function, cmd_args, self.namespace,
                                    self.arg_mode, debug=self.debug)
                print_result(result, output_mode)
                self.result = result
                sys.argv[:] # mark as accessed
            else:
                if not info:
                    # We guessed wrong earlier and didn't log yet; log now.
                    logger.info(cmd)
                print_result(result, output_mode)
                self.result = result
                unaccessed = sys.argv.unaccessed
                if unaccessed:
                    logger.error(
                        "%s is not callable.  Unexpected argument(s): %s",
                        result, " ".join(unaccessed))
                sys.argv[:] # don't complain again

    def run_module(self, module, args):
        arg_mode = _interpret_arg_mode(self.arg_mode, default="string")
        if arg_mode != "string":
            raise NotImplementedError(
                "run_module only supports string arguments")
        module = ModuleHandle(module)
        logger.info("python -m %s", ' '.join([str(module.name)] + args))
        # Imitate 'python -m'.
        # TODO: include only the traceback below runpy.run_module
        #   os.execvp(sys.executable, [sys.argv[0], "-m", modname] + args)
        sys.argv = [str(module.filename)] + args
        import runpy
        if self.debug:
            # TODO: break closer to user code
            debugger()
        try:
            runpy.run_module(str(module.name), run_name="__main__")
        except SystemExit:
            raise
        except:
            _handle_user_exception()

    def heuristic_run_module(self, module, args):
        module = ModuleHandle(module)
        # If the user ran 'py numpy --version', then print the numpy
        # version, i.e. same as 'py --version numpy'.  This has the
        # downside of shadowing a possible "--version" feature
        # implemented by the module itself.  However, this is probably
        # not a big deal, because (1) a full-featured program that
        # supports --version would normally have a driver script and
        # not rely on 'python -m foo'; (2) it would probably do
        # something similar anyway; (3) the user can do 'py -m foo
        # --version' if necessary.
        if len(args)==1 and args[0] in ["--version", "-version"]:
            self.print_version(module)
            return
        if len(args)==1 and args[0] in ["--help", "-help", "--h", "-h",
                                        "--?", "-?", "?"]:
            expr = UserExpr(module.module, None, "raw_value",
                            source=str(module.name))
            usage = _get_help(expr, 1)
            print(usage)
            return
        if len(args)==1 and args[0] in ["--source", "-source",
                                        "--??", "-??", "??"]:
            expr = UserExpr(module.module, None, "raw_value",
                            source=str(module.name))
            usage = _get_help(expr, 2)
            print(usage)
            return
        # TODO: check if module checks __main__
        self.run_module(module, args)

    def print_version(self, arg):
        if not arg:
            print_version_and_exit()
            return
        if isinstance(arg, (ModuleHandle, types.ModuleType)):
            module = ModuleHandle(arg).module
        else:
            module = self.namespace.auto_eval(arg, mode="eval")
        if not isinstance(module, types.ModuleType):
            raise TypeError("print_version(): got a %s instead of a module"
                            % (type(module).__name__,))
        try:
            version = module.__version__
        except AttributeError:
            raise AttributeError(
                "Module %s does not have a __version__ attribute"
                % (module.__name__,))
        print(version)

    def print_help(self, objname, verbosity=1):
        objname = objname and objname.strip()
        if not objname:
            print(__doc__)
            return
        expr = UserExpr(objname, self.namespace, "eval")
        usage = _get_help(expr, verbosity)
        print(usage)

    def create_ipython_app(self):
        """
        Create an IPython application and initialize it, but don't start it.
        """
        assert self.ipython_app is None
        self.ipython_app = get_ipython_terminal_app_with_autoimporter()

    def start_ipython(self, args=[]):
        user_ns = self.namespace.globals
        start_ipython_with_autoimporter(args, _user_ns=user_ns,
                                        app=self.ipython_app)
        # Don't need to do another interactive session after this one
        # (i.e. make 'py --interactive' the same as 'py').
        self.interactive = False

    def _parse_global_opts(self):
        args = list(self.main_args)
        self.add_deprecated_builtins = False
        self.debug       = False
        self.interactive = False
        self.verbosity   = 1
        self.arg_mode    = None
        self.output_mode = None
        postmortem = 'auto'
        while args:
            arg = args[0]
            if arg in ["debug", "pdb", "ipdb", "dbg"]:
                argname = "debug"
            elif not arg.startswith("-"):
                break
            elif arg.startswith("--"):
                argname = arg[2:]
            else:
                argname = arg[1:]
            argname, equalsign, value = argname.partition("=")
            def popvalue():
                if equalsign:
                    return value
                else:
                    try:
                        return args.pop(0)
                    except IndexError:
                        raise ValueError("expected argument to %s" % arg)
            def novalue():
                if equalsign:
                    raise ValueError("unexpected argument %s" % arg)
            if argname in ["interactive", "i"]:
                novalue()
                self.interactive = True
                del args[0]
                # Create and initialize the IPython app now (but don't start
                # it yet).  We'll use it later.  The reason to initialize it
                # now is that the code that we're running might check if it's
                # running in interactive mode based on whether an IPython app
                # has been initialized.  Some user code even initializes
                # things differently at module import time based on this.
                self.create_ipython_app()
            elif argname in ["debug", "pdb", "ipdb", "dbg", "d"]:
                novalue()
                self.debug = True
                del args[0]
            if argname == "verbose":
                novalue()
                logger.set_level("DEBUG")
                del args[0]
                continue
            if argname in ["quiet", "q"]:
                novalue()
                logger.set_level("ERROR")
                del args[0]
                continue
            if argname in ["safe"]:
                del args[0]
                novalue()
                self.arg_mode = _interpret_arg_mode("string")
                # TODO: make this less hacky, something like
                #   self.import_db = ""
                # TODO: also turn off which() behavior
                os.environ["PYFLYBY_PATH"] = "EMPTY"
                os.environ["PYFLYBY_KNOWN_IMPORTS_PATH"] = ""
                os.environ["PYFLYBY_MANDATORY_IMPORTS_PATH"] = ""
                continue
            if argname in ["arguments", "argument", "args", "arg",
                           "arg_mode", "arg-mode", "argmode"]:
                # Interpret --args=eval|string|auto.
                # Note that if the user didn't specify --args, then we
                # intentionally leave ``opts.arg_mode`` set to ``None`` for now,
                # because the default varies per action.
                del args[0]
                self.arg_mode = _interpret_arg_mode(popvalue())
                continue
            if argname in ["output", "output_mode", "output-mode",
                           "out", "outmode", "out_mode", "out-mode", "o"]:
                del args[0]
                self.output_mode = _interpret_output_mode(popvalue())
                continue
            if argname in ["print", "pprint", "silent", "repr"]:
                del args[0]
                novalue()
                self.output_mode = _interpret_output_mode(argname)
                continue
            if argname in ["postmortem"]:
                del args[0]
                v = (value or "").lower().strip()
                if v in ["yes", "y", "always", "true", "t", "1", "enable", ""]:
                    postmortem = True
                elif v in ["no", "n", "never", "false", "f", "0", "disable"]:
                    postmortem = False
                elif v in ["auto", "automatic", "default", "if-tty"]:
                    postmortem = "auto"
                else:
                    raise ValueError(
                        "unexpected %s=%s.  "
                        "Try --postmortem=yes or --postmortem=no."
                        % (argname, value))
                continue
            if argname in ["no-postmortem", "np"]:
                del args[0]
                novalue()
                postmortem = False
                continue
            if argname in ["add-deprecated-builtins", "add_deprecated_builtins"]:
                del args[0]
                self.add_deprecated_builtins = True
                continue
            break
        self.args = args
        if postmortem == "auto":
            postmortem = os.isatty(1)
        global _enable_postmortem_debugger
        _enable_postmortem_debugger = postmortem

    def _enable_debug_tools(self, *, add_deprecated: bool):
        # Enable a bunch of debugging tools.
        enable_faulthandler()
        enable_signal_handler_debugger()
        enable_sigterm_handler()
        add_debug_functions_to_builtins(add_deprecated=add_deprecated)

    def run(self):
        # Parse global options.
        sys.orig_argv = list(sys.argv)
        self._parse_global_opts()
        self._enable_debug_tools(add_deprecated=self.add_deprecated_builtins)
        self._run_action()
        self._pre_exit()

    def _run_action(self):
        args = self.args
        if not args or args[0] == "-":
            if os.isatty(0):
                # The user specified no arguments (or only a "-") and stdin is a
                # tty.  Run ipython with autoimporter enabled, i.e. equivalent to
                # autoipython.  Note that we directly start IPython in the same
                # process, instead of using subprocess.call(['autoipython']),
                # because the latter messes with Control-C handling.
                # TODO: add 'py shell' and make this an alias.
                # TODO: if IPython isn't installed, then do our own
                # interactive REPL with code.InteractiveConsole, readline, and
                # autoimporter.
                self.start_ipython()
                return
            else:
                # Emulate python args.
                cmd_args = args or [""]
                # Execute code from stdin, with auto importing.
                self.exec_stdin(cmd_args)
                return

        # Consider --action=arg, --action arg, -action=arg, -action arg,
        # %action arg.
        arg0 = args.pop(0)
        if not arg0.strip():
            raise ValueError("got empty string as first argument")
        if arg0.startswith("--"):
            action, equalsign, cmdarg = arg0[2:].partition("=")
        elif arg0.startswith("-"):
            action, equalsign, cmdarg = arg0[1:].partition("=")
        elif arg0.startswith("%"):
            action, equalsign, cmdarg = arg0[1:], None, None
        elif len(arg0) > 1 or arg0 == "?":
            action, equalsign, cmdarg = arg0, None, None
        else:
            action, equalsign, cmdarg = None, None, None
        def popcmdarg():
            if equalsign:
                return cmdarg
            else:
                try:
                    return args.pop(0)
                except IndexError:
                    raise ValueError("expected argument to %s" % arg0)
        def nocmdarg():
            if equalsign:
                raise ValueError("unexpected argument %s" % arg0)

        if action in ["eval", "c", "e"]:
            # Evaluate a command from the command-line, with auto importing.
            # Supports expressions as well as statements.
            # Note: Regular python supports "python -cfoo" as equivalent to
            # "python -c foo".  For now, we intentionally don't support that.
            cmd = popcmdarg()
            self.eval(cmd, args)
        elif action in ["file", "execfile", "execf", "runfile", "run", "f",
                        "python"]:
            # Execute a file named on the command-line, with auto importing.
            cmd = popcmdarg()
            self.execfile(cmd, args)
        elif action in ["apply", "call"]:
            # Call a function named on the command-line, with auto importing and
            # auto evaluation of arguments.
            function_name = popcmdarg()
            function = UserExpr(function_name, self.namespace, "eval")
            self.apply(function, args)
        elif action in ["map"]:
            # Call function on each argument.
            # TODO: instead of making this a standalone mode, change this to a
            # flag that can eval/apply/exec/etc.  Set "_" to each argument.
            # E.g. py --map 'print _' obj1 obj2
            #      py --map _.foo obj1 obj2
            #      py --map '_**2' 3 4 5
            # when using heuristic mode, "lock in" the action mode on the
            # first argument.
            function_name = popcmdarg()
            function = UserExpr(function_name, self.namespace, "eval")
            if args and args[0] == '--':
                for arg in args[1:]:
                    self.apply(function, ['--', arg])
            else:
                for arg in args:
                    self.apply(function, [arg])
        elif action in ["xargs"]:
            # TODO: read lines from stdin and map.  default arg_mode=string
            raise NotImplementedError("TODO: xargs")
        elif action in ["module", "m", "runmodule", "run_module", "run-module"]:
            # Exactly like `python -m'.  Intentionally does NOT do auto
            # importing within the module, because modules should not be
            # sloppy; they should instead be tidied to have the correct
            # imports.
            modname = popcmdarg()
            self.run_module(modname, args)
        elif arg0.startswith("-m"):
            # Support "py -mbase64" in addition to "py -m base64".
            modname = arg0[2:]
            self.run_module(modname, args)
        elif action in ["attach"]:
            pid = int(popcmdarg())
            nocmdarg()
            attach_debugger(pid)
        elif action in ["stack", "stack_trace", "stacktrace", "backtrace", "bt"]:
            pid = int(popcmdarg())
            nocmdarg()
            print("Stack trace for process %s:" % (pid,))
            remote_print_stack(pid)
        elif action in ["ipython", "ip"]:
            # Start IPython.
            self.start_ipython(args)
        elif action in ["notebook", "nb"]:
            # Start IPython notebook.
            nocmdarg()
            self.start_ipython(["notebook"] + args)
        elif action in ["kernel"]:
            # Start IPython kernel.
            nocmdarg()
            self.start_ipython(["kernel"] + args)
        elif action in ["qtconsole", "qt"]:
            # Start IPython qtconsole.
            nocmdarg()
            self.start_ipython(["qtconsole"] + args)
        elif action in ["console"]:
            # Start IPython console (with new kernel).
            nocmdarg()
            self.start_ipython(["console"] + args)
        elif action in ["existing"]:
            # Start IPython console (with existing kernel).
            if equalsign:
                args.insert(0, cmdarg)
            self.start_ipython(["console", "--existing"] + args)
        elif action in ["nbconvert"]:
            # Start IPython nbconvert.  (autoimporter is irrelevant.)
            if equalsign:
                args.insert(0, cmdarg)
            start_ipython_with_autoimporter(["nbconvert"] + args)
        elif action in ["timeit"]:
            # TODO: make --timeit and --time flags which work with any mode
            # and heuristic, instead of only eval.
            # TODO: fallback if IPython isn't available.  above todo probably
            # requires not using IPython anyway.
            nocmdarg()
            run_ipython_line_magic("%timeit " + ' '.join(args))
        elif action in ["time"]:
            # TODO: make --timeit and --time flags which work with any mode
            # and heuristic, instead of only eval.
            # TODO: fallback if IPython isn't available.  above todo probably
            # requires not using IPython anyway.
            nocmdarg()
            run_ipython_line_magic("%time " + ' '.join(args))
        elif action in ["version"]:
            if equalsign:
                args.insert(0, cmdarg)
            self.print_version(args[0] if args else None)
        elif action in ["help", "h", "?"]:
            if equalsign:
                args.insert(0, cmdarg)
            self.print_help(args[0] if args else None, verbosity=1)
        elif action in ["pinfo"]:
            self.print_help(popcmdarg(), verbosity=1)
        elif action in ["source", "pinfo2", "??"]:
            self.print_help(popcmdarg(), verbosity=2)

        elif arg0.startswith("-"):
            # Unknown argument.
            msg = "Unknown option %s" % (arg0,)
            if arg0.startswith("-c"):
                msg += "; do you mean -c %s?" % (arg0[2:])
            syntax(msg, usage=usage)

        elif arg0.startswith("??"):
            # TODO: check number of args
            self.print_help(arg0[2:], verbosity=2)

        elif arg0.endswith("??"):
            # TODO: check number of args
            self.print_help(arg0[:-2], verbosity=2)

        elif arg0.startswith("?"):
            # TODO: check number of args
            self.print_help(arg0[1:], verbosity=1)

        elif arg0.endswith("?"):
            # TODO: check number of args
            self.print_help(arg0[:-1], verbosity=1)

        elif arg0.startswith("%"):
            run_ipython_line_magic(' '.join([arg0]+args))

        # Heuristically choose the behavior automatically based on what the
        # argument looks like.
        else:
            filename = _as_filename_if_seems_like_filename(arg0)
            if filename:
                # Implied --execfile.
                self.execfile(filename, args)
                return
            if not args and arg0.isdigit():
                if self.debug:
                    attach_debugger(int(arg0, 10))
                    return
                else:
                    logger.error(
                        "Use py -d %s if you want to attach a debugger", arg0)
                    raise SystemExit(1)
            # Implied --eval.
            # When given multiple arguments, first see if the args can be
            # concatenated and parsed as a single python program/expression.
            # But don't try this if any arguments look like options, empty
            # string or whitespace, etc.
            # TODO: refactor
            if (args and
                self.arg_mode == None and
                not any(re.match(r"\s*$|-[a-zA-Z-]", a) for a in args)):
                cmd = PythonBlock(" ".join([arg0]+args),
                                  flags=FLAGS, auto_flags=True)
                if cmd.parsable and self.namespace.auto_import(cmd):
                    with SysArgvCtx("-c"):
                        output_mode = _interpret_output_mode(self.output_mode)
                        result = self.namespace.auto_eval(
                            cmd, info=True, auto_import=False)
                        print_result(result, output_mode)
                        self.result = result
                        return
                # else fall through
            # Heuristic based on first arg: try run_module, apply, or exec/eval.
            cmd = PythonBlock(arg0, flags=FLAGS, auto_flags=True)
            if not cmd.parsable:
                logger.error(
                    "Could not interpret as filename or expression: %s",
                    arg0)
                syntax(usage=usage)
            self.heuristic_cmd(cmd, args, function_name=arg0)

    def _pre_exit(self):
        self._pre_exit_matplotlib_show()
        self._pre_exit_interactive_shell()

    def _pre_exit_matplotlib_show(self):
        """
        If matplotlib.pyplot (pylab) is loaded, then call the show() function.
        This will cause the program to block until all figures are closed.
        Without this, a command like 'py plot(...)' would exit immediately.
        """
        if self.interactive:
            return
        try:
            pyplot = sys.modules["matplotlib.pyplot"]
        except KeyError:
            return
        pyplot.show()

    def _pre_exit_interactive_shell(self):
        if self.interactive:
            assert self.ipython_app is not None
            self.namespace.globals["_"] = self.result
            self.start_ipython()


def py_main(args=None):
    if args is None:
        args = sys.argv[1:]
    _PyMain(args).run()



# pyflyby/_util.py.
# Copyright (C) 2011, 2012, 2013, 2014, 2015, 2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT



from   contextlib               import contextmanager, ExitStack
import inspect
import os
import sys
import types

# There used to be a custom caching_attribute implementation
# this now uses functools's cached_property which is understood by
# various static analysis tools.
from functools import cached_property as cached_attribute # noqa: F401

# Python 2/3 compatibility
DictProxyType = type(object.__dict__)

if sys.version_info > (3,9):
    from functools import cache as memoize
else:
    def memoize(function):
        cache = {}
        def wrapped_fn(*args, **kwargs):
            cache_key = (args, tuple(sorted(kwargs.items())))
            try:
                return cache[cache_key]
            except KeyError:
                result = function(*args, **kwargs)
                cache[cache_key] = result
                return result
        wrapped_fn.cache = cache
        return wrapped_fn


class WrappedAttributeError(Exception):
    pass


def stable_unique(items):
    """
    Return a copy of ``items`` without duplicates.  The order of other items is
    unchanged.

      >>> stable_unique([1,4,6,4,6,5,7])
      [1, 4, 6, 5, 7]
    """
    result = []
    seen = set()
    for item in items:
        if item in seen:
            continue
        seen.add(item)
        result.append(item)
    return result


def longest_common_prefix(items1, items2):
    """
    Return the longest common prefix.

      >>> longest_common_prefix("abcde", "abcxy")
      'abc'

    :rtype:
      ``type(items1)``
    """
    n = 0
    for x1, x2 in zip(items1, items2):
        if x1 != x2:
            break
        n += 1
    return items1[:n]


def prefixes(parts):
    """
      >>> list(prefixes("abcd"))
      ['a', 'ab', 'abc', 'abcd']

    """
    for i in range(1, len(parts)+1):
        yield parts[:i]


def indent(lines, prefix):
    r"""
      >>> indent('hello\nworld\n', '@@')
      '@@hello\n@@world\n'
    """
    return "".join("%s%s\n"%(prefix,line) for line in lines.splitlines(False))


def partition(iterable, predicate):
    """
      >>> partition('12321233221', lambda c: int(c) % 2 == 0)
      (['2', '2', '2', '2', '2'], ['1', '3', '1', '3', '3', '1'])

    """
    falses = []
    trues = []
    for item in iterable:
        if predicate(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses


Inf = float('Inf')


@contextmanager
def NullCtx():
    """
    Context manager that does nothing.
    """
    yield


@contextmanager
def ImportPathCtx(path_additions):
    """
    Context manager that temporarily prepends ``sys.path`` with ``path_additions``.
    """
    if not isinstance(path_additions, (tuple, list)):
        path_additions = [path_additions]
    old_path = sys.path[:]
    sys.path[0:0] = path_additions
    try:
        yield
    finally:
        sys.path[:] = old_path


@contextmanager
def CwdCtx(path):
    """
    Context manager that temporarily enters a new working directory.
    """
    old_cwd = os.getcwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(old_cwd)


@contextmanager
def EnvVarCtx(**kwargs):
    """
    Context manager that temporarily modifies os.environ.
    """
    unset = object()
    old = {}
    try:
        for k, v in kwargs.items():
            old[k] = os.environ.get(k, unset)
            os.environ[k] = v
        yield
    finally:
        for k, v in old.items():
            if v is unset:
                del os.environ[k]
            else:
                os.environ[k] = v


@contextmanager
def ExcludeImplicitCwdFromPathCtx():
    """
    Context manager that temporarily removes "." from ``sys.path``.
    """
    old_path = sys.path
    try:
        sys.path = [p for p in sys.path if p not in (".", "")]
        yield
    finally:
        sys.path[:] = old_path


class FunctionWithGlobals(object):
    """
    A callable that at runtime adds extra variables to the target function's
    global namespace.

    This is written as a class with a __call__ method.  We do so rather than
    using a metafunction, so that we can also implement __getattr__ to look
    through to the target.
    """

    def __init__(self, function, **variables):
        self.__function = function
        self.__variables = variables
        try:
            self.__original__ = variables["__original__"]
        except KeyError:
            pass

    def __call__(self, *args, **kwargs):
        function = self.__function
        variables = self.__variables
        undecorated = function
        while True:
            try:
                undecorated = undecorated.undecorated
            except AttributeError:
                break
        globals = undecorated.__globals__
        UNSET = object()
        old = {}
        for k in variables:
            old[k] = globals.get(k, UNSET)
        try:
            for k, v in variables.items():
                globals[k] = v
            return function(*args, **kwargs)
        finally:
            for k, v in old.items():
                if v is UNSET:
                    del globals[k]
                else:
                    globals[k] = v

    def __getattr__(self, k):
        return getattr(self.__original__, k)


    def __get__(self, inst, cls=None):
        if inst is None:
            return self
        return types.MethodType(self, inst)



class _WritableDictProxy(object):
    """
    Writable equivalent of cls.__dict__.
    """

    # We need to implement __getitem__ differently from __setitem__.  The
    # reason is because of an asymmetry in the mechanics of classes:
    #   - getattr(cls, k) does NOT in general do what we want because it
    #     returns unbound methods.  It's actually equivalent to
    #     cls.__dict__[k].__get__(cls).
    #   - setattr(cls, k, v) does do what we want.
    #   - cls.__dict__[k] does do what we want.
    #   - cls.__dict__[k] = v does not work, because dictproxy is read-only.

    def __init__(self, cls):
        self._cls = cls

    def __getitem__(self, k):
        return self._cls.__dict__[k]

    def get(self, k, default=None):
        return self._cls.__dict__.get(k, default)

    def __setitem__(self, k, v):
        setattr(self._cls, k, v)

    def __delitem__(self, k):
        delattr(self._cls, k)


_UNSET = object()

class Aspect(object):
    """
    Monkey-patch a target method (joinpoint) with "around" advice.

    The advice can call "__original__(...)".  At run time, a global named
    "__original__" will magically be available to the wrapped function.
    This refers to the original function.

    Suppose someone else wrote Foo.bar()::

      >>> class Foo(object):
      ...     def __init__(self, x):
      ...         self.x = x
      ...     def bar(self, y):
      ...         return "bar(self.x=%s,y=%s)" % (self.x,y)

      >>> foo = Foo(42)

    To monkey patch ``foo.bar``, decorate the wrapper with ``"@advise(foo.bar)"``::

      >>> @advise(foo.bar)
      ... def addthousand(y):
      ...     return "advised foo.bar(y=%s): %s" % (y, __original__(y+1000))

      >>> foo.bar(100)
      'advised foo.bar(y=100): bar(self.x=42,y=1100)'

    You can uninstall the advice and get the original behavior back::

      >>> addthousand.unadvise()

      >>> foo.bar(100)
      'bar(self.x=42,y=100)'

    :see:
      http://en.wikipedia.org/wiki/Aspect-oriented_programming
    """

    _wrapper = None

    def __init__(self, joinpoint):
        spec = joinpoint
        while hasattr(joinpoint, "__joinpoint__"):
            joinpoint = joinpoint.__joinpoint__
        self._joinpoint = joinpoint
        if (isinstance(joinpoint, (types.FunctionType, type))
            and not (joinpoint.__name__ != joinpoint.__qualname__)):
            self._qname = "%s.%s" % (
                joinpoint.__module__,
                joinpoint.__name__)
            self._container = sys.modules[joinpoint.__module__].__dict__
            self._name      = joinpoint.__name__
            self._original  = spec
            assert spec == self._container[self._name], joinpoint
        elif isinstance(joinpoint, types.MethodType) or (isinstance(joinpoint,
            types.FunctionType) and joinpoint.__name__ !=
            joinpoint.__qualname__) or isinstance(joinpoint, property):
            if isinstance(joinpoint, property):
                joinpoint = joinpoint.fget
                self._wrapper = property
            self._qname = '%s.%s' % (joinpoint.__module__,
                                     joinpoint.__qualname__)
            self._name      = joinpoint.__name__
            if getattr(joinpoint, '__self__', None) is None:
                container_obj   = getattr(inspect.getmodule(joinpoint),
                   joinpoint.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])

                self._container = _WritableDictProxy(container_obj)
                self._original  = spec
            else:
                # Instance method.
                container_obj   = joinpoint.__self__
                self._container = container_obj.__dict__
                self._original  = spec
            assert spec == getattr(container_obj, self._name), (container_obj, self._qname)
            assert self._original == self._container.get(self._name, self._original)
        elif isinstance(joinpoint, tuple) and len(joinpoint) == 2:
            container, name = joinpoint
            if isinstance(container, dict):
                self._original  = container[name]
                self._container = container
                self._qname = name
            elif name in container.__dict__.get('_trait_values', ()):
                # traitlet stuff from IPython
                self._container = container._trait_values
                self._original = self._container[name]
                self._qname = name
            elif isinstance(container.__dict__, DictProxyType):
                original = getattr(container, name)
                if hasattr(original, "__func__"):
                    # TODO: generalize this to work for all cases, not just classmethod
                    original = original.__func__
                    self._wrapper = classmethod
                self._original = original
                self._container = _WritableDictProxy(container)
                self._qname = "%s.%s.%s" % (
                    container.__module__, container.__name__, name)
            else:
                # Keep track of the original.  We use getattr on the
                # container, instead of getitem on container.__dict__, so that
                # it works even if it's a class dict proxy that inherits the
                # value from a super class.
                self._original = getattr(container, name)
                self._container = container.__dict__
                self._qname = "%s.%s.%s" % (
                    container.__class__.__module__,
                    container.__class__.__name__,
                    name)
            self._name      = name
        # TODO: unbound method
        else:
            raise TypeError("JoinPoint: unexpected type %s"
                            % (type(joinpoint).__name__,))
        self._wrapped = None

    def advise(self, hook, once=False):
        from pyflyby._log import logger
        self._previous = self._container.get(self._name, _UNSET)
        if once and getattr(self._previous, "__aspect__", None) :
            # TODO: check that it's the same hook - at least check the name.
            logger.debug("already advised %s", self._qname)
            return None
        logger.debug("advising %s", self._qname)
        assert self._previous is _UNSET or self._previous == self._original
        assert self._wrapped is None
        # Create the wrapped function.
        wrapped = FunctionWithGlobals(hook, __original__=self._original)
        wrapped.__joinpoint__ = self._joinpoint
        wrapped.__original__ = self._original
        wrapped.__name__ = "%s__advised__%s" % (self._name, hook.__name__)
        wrapped.__doc__ = "%s.\n\nAdvice %s:\n%s" % (
            self._original.__doc__, hook.__name__, hook.__doc__)
        wrapped.__aspect__ = self
        if self._wrapper is not None:
            wrapped = self._wrapper(wrapped)
        self._wrapped = wrapped
        # Install the wrapped function!
        self._container[self._name] = wrapped
        return self

    def unadvise(self):
        if self._wrapped is None:
            return
        cur = self._container.get(self._name, _UNSET)
        if cur is self._wrapped:
            from pyflyby._log import logger
            logger.debug("unadvising %s", self._qname)
            if self._previous is _UNSET:
                del self._container[self._name]
            else:
                self._container[self._name] = self._previous
        elif cur == self._previous:
            pass
        else:
            from pyflyby._log import logger
            logger.debug("%s seems modified; not unadvising it", self._name)
        self._wrapped = None


def advise(joinpoint):
    """
    Advise ``joinpoint``.

    See `Aspect`.
    """
    aspect = Aspect(joinpoint)
    return aspect.advise


@contextmanager
def AdviceCtx(joinpoint, hook):
    aspect = Aspect(joinpoint)
    advice = aspect.advise(hook)
    try:
        yield
    finally:
        advice.unadvise()

# For Python 2/3 compatibility. cmp isn't included with six.
def cmp(a, b):
    return (a > b) - (a < b)


# Create a context manager with an arbitrary number of contexts.
@contextmanager
def nested(*mgrs):
    with ExitStack() as stack:
        ctxes = [stack.enter_context(mgr) for mgr in mgrs]
        yield ctxes
